# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic actor implementation, using TensorFlow and Sonnet."""

from typing import Optional, Tuple

from acme import adders
from acme import core
from acme import types
# Internal imports.
from acme.tf import utils as tf2_utils
from acme.tf import variable_utils as tf2_variable_utils

import dm_env
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp

tfd = tfp.distributions


class FeedForwardActor(core.Actor):
  """A feed-forward actor.

  An actor based on a feed-forward policy which takes non-batched observations
  and outputs non-batched actions. It also allows adding experiences to replay
  and updating the weights from the policy on the learner.
  """

  def __init__(
      self,
      policy_network: snt.Module,
      adder: Optional[adders.Adder] = None,
      variable_client: Optional[tf2_variable_utils.VariableClient] = None,
  ):
    """Initializes the actor.

    Args:
      policy_network: the policy to run.
      adder: the adder object to which allows to add experiences to a
        dataset/replay buffer.
      variable_client: object which allows to copy weights from the learner copy
        of the policy to the actor copy (in case they are separate).
    """

    # Store these for later use.
    self._adder = adder
    self._variable_client = variable_client
    self._policy_network = policy_network

  @tf.function
  def _policy(self, observation: types.NestedTensor) -> types.NestedTensor:
    # Add a dummy batch dimension and as a side effect convert numpy to TF.
    batched_observation = tf2_utils.add_batch_dim(observation)

    # Compute the policy, conditioned on the observation.
    policy = self._policy_network(batched_observation)

    # Sample from the policy if it is stochastic.
    action = policy.sample() if isinstance(policy, tfd.Distribution) else policy

    return action

  def select_action(self, observation: types.NestedArray) -> types.NestedArray:
    # Pass the observation through the policy network.
    action = self._policy(observation)

    # Return a numpy array with squeezed out batch dimension.
    return tf2_utils.to_numpy_squeeze(action)

  def observe_first(self, timestep: dm_env.TimeStep):
    if self._adder:
      self._adder.add_first(timestep)

  def observe(self, action: types.NestedArray, next_timestep: dm_env.TimeStep):
    if self._adder:
      self._adder.add(action, next_timestep)

  def update(self, wait: bool = False):
    if self._variable_client:
      self._variable_client.update(wait)


class RecurrentActor(core.Actor):
  """A recurrent actor.

  An actor based on a recurrent policy which takes non-batched observations and
  outputs non-batched actions, and keeps track of the recurrent state inside. It
  also allows adding experiences to replay and updating the weights from the
  policy on the learner.
  """

  def __init__(
      self,
      policy_network: snt.RNNCore,
      adder: Optional[adders.Adder] = None,
      variable_client: Optional[tf2_variable_utils.VariableClient] = None,
      store_recurrent_state: bool = True,
  ):
    """Initializes the actor.

    Args:
      policy_network: the (recurrent) policy to run.
      adder: the adder object to which allows to add experiences to a
        dataset/replay buffer.
      variable_client: object which allows to copy weights from the learner copy
        of the policy to the actor copy (in case they are separate).
      store_recurrent_state: Whether to pass the recurrent state to the adder.
    """
    # Store these for later use.
    self._adder = adder
    self._variable_client = variable_client
    self._network = policy_network
    self._state = None
    self._prev_state = None
    self._store_recurrent_state = store_recurrent_state

  @tf.function
  def _policy(
      self,
      observation: types.NestedTensor,
      state: types.NestedTensor,
  ) -> Tuple[types.NestedTensor, types.NestedTensor]:

    # Add a dummy batch dimension and as a side effect convert numpy to TF.
    batched_observation = tf2_utils.add_batch_dim(observation)

    # Compute the policy, conditioned on the observation.
    policy, new_state = self._network(batched_observation, state)

    # Sample from the policy if it is stochastic.
    action = policy.sample() if isinstance(policy, tfd.Distribution) else policy

    return action, new_state

  def select_action(self, observation: types.NestedArray) -> types.NestedArray:
    # Initialize the RNN state if necessary.
    if self._state is None:
      self._state = self._network.initial_state(1)

    # Step the recurrent policy forward given the current observation and state.
    policy_output, new_state = self._policy(observation, self._state)

    # Bookkeeping of recurrent states for the observe method.
    self._prev_state = self._state
    self._state = new_state

    # Return a numpy array with squeezed out batch dimension.
    return tf2_utils.to_numpy_squeeze(policy_output)

  def observe_first(self, timestep: dm_env.TimeStep):
    if self._adder:
      self._adder.add_first(timestep)

    # Set the state to None so that we re-initialize at the next policy call.
    self._state = None

  def observe(self, action: types.NestedArray, next_timestep: dm_env.TimeStep):
    if not self._adder:
      return

    if not self._store_recurrent_state:
      self._adder.add(action, next_timestep)
      return

    numpy_state = tf2_utils.to_numpy_squeeze(self._prev_state)
    self._adder.add(action, next_timestep, extras=(numpy_state,))

  def update(self, wait: bool = False):
    if self._variable_client:
      self._variable_client.update(wait)


# Internal class 1.
# Internal class 2.



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transition adders.

This implements an N-step transition adder which collapses trajectory sequences
into a single transition, simplifying to a simple transition adder when N=1.
"""

import copy
from typing import Optional, Tuple

from acme import specs
from acme import types
from acme.adders.reverb import base
from acme.adders.reverb import utils
from acme.utils import tree_utils

import numpy as np
import reverb
import tree


class NStepTransitionAdder(base.ReverbAdder):
  """An N-step transition adder.

  This will buffer a sequence of N timesteps in order to form a single N-step
  transition which is added to reverb for future retrieval.

  For N=1 the data added to replay will be a standard one-step transition which
  takes the form:

        (s_t, a_t, r_t, d_t, s_{t+1}, e_t)

  where:

    s_t = state observation at time t
    a_t = the action taken from s_t
    r_t = reward ensuing from action a_t
    d_t = environment discount ensuing from action a_t. This discount is
        applied to future rewards after r_t.
    e_t [Optional] = extra data that the agent persists in replay.

  For N greater than 1, transitions are of the form:

        (s_t, a_t, R_{t:t+n}, D_{t:t+n}, s_{t+N}, e_t),

  where:

    s_t = State (observation) at time t.
    a_t = Action taken from state s_t.
    g = the additional discount, used by the agent to discount future returns.
    R_{t:t+n} = N-step discounted return, i.e. accumulated over N rewards:
          R_{t:t+n} := r_t + g * d_t * r_{t+1} + ...
                           + g^{n-1} * d_t * ... * d_{t+n-2} * r_{t+n-1}.
    D_{t:t+n}: N-step product of agent discounts g_i and environment
      "discounts" d_i.
          D_{t:t+n} := g^{n-1} * d_{t} * ... * d_{t+n-1},
      For most environments d_i is 1 for all steps except the last,
      i.e. it is the episode termination signal.
    s_{t+n}: The "arrival" state, i.e. the state at time t+n.
    e_t [Optional]: A nested structure of any 'extras' the user wishes to add.

  Notes:
    - At the beginning and end of episodes, shorter transitions are added.
      That is, at the beginning of the episode, it will add:
            (s_0 -> s_1), (s_0 -> s_2), ..., (s_0 -> s_n), (s_1 -> s_{n+1})

      And at the end of the episode, it will add:
            (s_{T-n+1} -> s_T), (s_{T-n+2} -> s_T), ... (s_{T-1} -> s_T).
    - We add the *first* `extra` of each transition, not the *last*, i.e.
        if extras are provided, we get e_t, not e_{t+n}.
  """

  def __init__(
      self,
      client: reverb.Client,
      n_step: int,
      discount: float,
      *,
      priority_fns: Optional[base.PriorityFnMapping] = None,
      max_in_flight_items: int = 5,
  ):
    """Creates an N-step transition adder.

    Args:
      client: A `reverb.Client` to send the data to replay through.
      n_step: The "N" in N-step transition. See the class docstring for the
        precise definition of what an N-step transition is. `n_step` must be at
        least 1, in which case we use the standard one-step transition, i.e.
        (s_t, a_t, r_t, d_t, s_t+1, e_t).
      discount: Discount factor to apply. This corresponds to the agent's
        discount in the class docstring.
      priority_fns: See docstring for BaseAdder.
      max_in_flight_items: The maximum number of items allowed to be "in flight"
        at the same time. See `block_until_num_items` in
        `reverb.TrajectoryWriter.flush` for more info.

    Raises:
      ValueError: If n_step is less than 1.
    """
    # Makes the additional discount a float32, which means that it will be
    # upcast if rewards/discounts are float64 and left alone otherwise.
    self.n_step = n_step
    self._discount = tree.map_structure(np.float32, discount)
    self._first_idx = 0
    self._last_idx = 0

    super().__init__(
        client=client,
        max_sequence_length=n_step + 1,
        priority_fns=priority_fns,
        max_in_flight_items=max_in_flight_items)

  def add(self, *args, **kwargs):
    # Increment the indices for the start and end of the window for computing
    # n-step returns.
    if self._writer.episode_steps >= self.n_step:
      self._first_idx += 1
    self._last_idx += 1

    super().add(*args, **kwargs)

  def reset(self):
    super().reset()
    self._first_idx = 0
    self._last_idx = 0

  @property
  def _n_step(self) -> int:
    """Effective n-step, which may vary at starts and ends of episodes."""
    return self._last_idx - self._first_idx

  def _write(self):
    # Convenient getters for use in tree operations.
    get_first = lambda x: x[self._first_idx]
    get_last = lambda x: x[self._last_idx]
    # Note: this getter is meant to be used on a TrajectoryWriter.history to
    # obtain its numpy values.
    get_all_np = lambda x: x[self._first_idx:self._last_idx].numpy()

    # Get the state, action, next_state, as well as possibly extras for the
    # transition that is about to be written.
    history = self._writer.history
    s, a = tree.map_structure(get_first,
                              (history['observation'], history['action']))
    s_ = tree.map_structure(get_last, history['observation'])

    # Maybe get extras to add to the transition later.
    if 'extras' in history:
      extras = tree.map_structure(get_first, history['extras'])

    # Note: at the beginning of an episode we will add the initial N-1
    # transitions (of size 1, 2, ...) and at the end of an episode (when
    # called from write_last) we will write the final transitions of size (N,
    # N-1, ...). See the Note in the docstring.
    # Get numpy view of the steps to be fed into the priority functions.
    reward, discount = tree.map_structure(
        get_all_np, (history['reward'], history['discount']))

    # Compute discounted return and geometric discount over n steps.
    n_step_return, total_discount = self._compute_cumulative_quantities(
        reward, discount)

    # Append the computed n-step return and total discount.
    # Note: if this call to _write() is within a call to _write_last(), then
    # this is the only data being appended and so it is not a partial append.
    self._writer.append(
        dict(n_step_return=n_step_return, total_discount=total_discount),
        partial_step=self._writer.episode_steps <= self._last_idx)
    # This should be done immediately after self._writer.append so the history
    # includes the recently appended data.
    history = self._writer.history

    # Form the n-step transition by using the following:
    # the first observation and action in the buffer, along with the cumulative
    # reward and discount computed above.
    n_step_return, total_discount = tree.map_structure(
        lambda x: x[-1], (history['n_step_return'], history['total_discount']))
    transition = types.Transition(
        observation=s,
        action=a,
        reward=n_step_return,
        discount=total_discount,
        next_observation=s_,
        extras=(extras if 'extras' in history else ()))

    # Calculate the priority for this transition.
    table_priorities = utils.calculate_priorities(self._priority_fns,
                                                  transition)

    # Insert the transition into replay along with its priority.
    for table, priority in table_priorities.items():
      self._writer.create_item(
          table=table, priority=priority, trajectory=transition)
      self._writer.flush(self._max_in_flight_items)

  def _write_last(self):
    # Write the remaining shorter transitions by alternating writing and
    # incrementingfirst_idx. Note that last_idx will no longer be incremented
    # once we're in this method's scope.
    self._first_idx += 1
    while self._first_idx < self._last_idx:
      self._write()
      self._first_idx += 1

  def _compute_cumulative_quantities(
      self, rewards: types.NestedArray, discounts: types.NestedArray
  ) -> Tuple[types.NestedArray, types.NestedArray]:

    # Give the same tree structure to the n-step return accumulator,
    # n-step discount accumulator, and self.discount, so that they can be
    # iterated in parallel using tree.map_structure.
    rewards, discounts, self_discount = tree_utils.broadcast_structures(
        rewards, discounts, self._discount)
    flat_rewards = tree.flatten(rewards)
    flat_discounts = tree.flatten(discounts)
    flat_self_discount = tree.flatten(self_discount)

    # Copy total_discount as it is otherwise read-only.
    total_discount = [np.copy(a[0]) for a in flat_discounts]

    # Broadcast n_step_return to have the broadcasted shape of
    # reward * discount.
    n_step_return = [
        np.copy(np.broadcast_to(r[0],
                                np.broadcast(r[0], d).shape))
        for r, d in zip(flat_rewards, total_discount)
    ]

    # NOTE: total_discount will have one less self_discount applied to it than
    # the value of self._n_step. This is so that when the learner/update uses
    # an additional discount we don't apply it twice. Inside the following loop
    # we will apply this right before summing up the n_step_return.
    for i in range(1, self._n_step):
      for nsr, td, r, d, sd in zip(n_step_return, total_discount, flat_rewards,
                                   flat_discounts, flat_self_discount):
        # Equivalent to: `total_discount *= self._discount`.
        td *= sd
        # Equivalent to: `n_step_return += reward[i] * total_discount`.
        nsr += r[i] * td
        # Equivalent to: `total_discount *= discount[i]`.
        td *= d[i]

    n_step_return = tree.unflatten_as(rewards, n_step_return)
    total_discount = tree.unflatten_as(rewards, total_discount)
    return n_step_return, total_discount

  # TODO(bshahr): make this into a standalone method. Class methods should be
  # used as alternative constructors or when modifying some global state,
  # neither of which is done here.
  @classmethod
  def signature(cls,
                environment_spec: specs.EnvironmentSpec,
                extras_spec: types.NestedSpec = ()):

    # This function currently assumes that self._discount is a scalar.
    # If it ever becomes a nested structure and/or a np.ndarray, this method
    # will need to know its structure / shape. This is because the signature
    # discount shape is the environment's discount shape and this adder's
    # discount shape broadcasted together. Also, the reward shape is this
    # signature discount shape broadcasted together with the environment
    # reward shape. As long as self._discount is a scalar, it will not affect
    # either the signature discount shape nor the signature reward shape, so we
    # can ignore it.

    rewards_spec, step_discounts_spec = tree_utils.broadcast_structures(
        environment_spec.rewards, environment_spec.discounts)
    rewards_spec = tree.map_structure(_broadcast_specs, rewards_spec,
                                      step_discounts_spec)
    step_discounts_spec = tree.map_structure(copy.deepcopy, step_discounts_spec)

    transition_spec = types.Transition(
        environment_spec.observations,
        environment_spec.actions,
        rewards_spec,
        step_discounts_spec,
        environment_spec.observations,  # next_observation
        extras_spec)

    return tree.map_structure_with_path(base.spec_like_to_tensor_spec,
                                        transition_spec)


def _broadcast_specs(*args: specs.Array) -> specs.Array:
  """Like np.broadcast, but for specs.Array.

  Args:
    *args: one or more specs.Array instances.

  Returns:
    A specs.Array with the broadcasted shape and dtype of the specs in *args.
  """
  bc_info = np.broadcast(*tuple(a.generate_value() for a in args))
  dtype = np.result_type(*tuple(a.dtype for a in args))
  return specs.Array(shape=bc_info.shape, dtype=dtype)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A iterator that does zero-copy conversion of `tf.Tensor`s into `np.ndarray`s."""

from typing import Iterator

from acme import types
import numpy as np
import tree


class NumpyIterator(Iterator[types.NestedArray]):
  """Iterator over a dataset with elements converted to numpy.

  Note: This iterator returns read-only numpy arrays.

  This iterator (compared to `tf.data.Dataset.as_numpy_iterator()`) does not
  copy the data when comverting `tf.Tensor`s to `np.ndarray`s.

  TODO(b/178684359): Remove this when it is upstreamed into `tf.data`.
  """

  __slots__ = ['_iterator']

  def __init__(self, dataset):
    self._iterator: Iterator[types.NestedTensor] = iter(dataset)

  def __iter__(self) -> 'NumpyIterator':
    return self

  def __next__(self) -> types.NestedArray:
    return tree.map_structure(lambda t: np.asarray(memoryview(t)),
                              next(self._iterator))

  def next(self):
    return self.__next__()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for acme.datasets.numpy_iterator."""

import collections

from absl.testing import absltest
from acme.datasets import numpy_iterator
import tensorflow as tf


class NumpyIteratorTest(absltest.TestCase):

  def testBasic(self):
    ds = tf.data.Dataset.range(3)
    self.assertEqual([0, 1, 2], list(numpy_iterator.NumpyIterator(ds)))

  def testNestedStructure(self):
    point = collections.namedtuple('Point', ['x', 'y'])
    ds = tf.data.Dataset.from_tensor_slices({
        'a': ([1, 2], [3, 4]),
        'b': [5, 6],
        'c': point([7, 8], [9, 10])
    })
    self.assertEqual([{
        'a': (1, 3),
        'b': 5,
        'c': point(7, 9)
    }, {
        'a': (2, 4),
        'b': 6,
        'c': point(8, 10)
    }], list(numpy_iterator.NumpyIterator(ds)))

if __name__ == '__main__':
  absltest.main()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for csv logging."""

import csv
import os

from absl.testing import absltest
from absl.testing import parameterized
from acme.testing import test_utils
from acme.utils import paths
from acme.utils.loggers import csv as csv_logger

_TEST_INPUTS = [{
    'c': 'foo',
    'a': '1337',
    'b': '42.0001',
}, {
    'c': 'foo2',
    'a': '1338',
    'b': '43.0001',
}]


class CSVLoggingTest(test_utils.TestCase):

  def test_logging_input_is_directory(self):

    # Set up logger.
    directory = self.get_tempdir()
    label = 'test'
    logger = csv_logger.CSVLogger(directory_or_file=directory, label=label)

    # Write data and close.
    for inp in _TEST_INPUTS:
      logger.write(inp)
    logger.close()

    # Read back data.
    outputs = []
    with open(logger.file_path) as f:
      csv_reader = csv.DictReader(f)
      for row in csv_reader:
        outputs.append(dict(row))
    self.assertEqual(outputs, _TEST_INPUTS)

  @parameterized.parameters(True, False)
  def test_logging_input_is_file(self, add_uid: bool):

    # Set up logger.
    directory = paths.process_path(
        self.get_tempdir(), 'logs', 'my_label', add_uid=add_uid)
    file = open(os.path.join(directory, 'logs.csv'), 'a')
    logger = csv_logger.CSVLogger(directory_or_file=file, add_uid=add_uid)

    # Write data and close.
    for inp in _TEST_INPUTS:
      logger.write(inp)
    logger.close()

    # Logger doesn't close the file; caller must do this manually.
    self.assertFalse(file.closed)
    file.close()

    # Read back data.
    outputs = []
    with open(logger.file_path) as f:
      csv_reader = csv.DictReader(f)
      for row in csv_reader:
        outputs.append(dict(row))
    self.assertEqual(outputs, _TEST_INPUTS)

  def test_flush(self):

    logger = csv_logger.CSVLogger(self.get_tempdir(), flush_every=1)
    for inp in _TEST_INPUTS:
      logger.write(inp)

    # Read back data.
    outputs = []
    with open(logger.file_path) as f:
      csv_reader = csv.DictReader(f)
      for row in csv_reader:
        outputs.append(dict(row))
    self.assertEqual(outputs, _TEST_INPUTS)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""IMPALA config."""
import dataclasses
from typing import Optional, Union

from acme import types
from acme.adders import reverb as adders_reverb
import numpy as np


@dataclasses.dataclass
class IMPALAConfig:
  """Configuration options for IMPALA."""
  seed: int = 0

  # Loss options
  batch_size: int = 16
  prefetch_size: int = 2
  sequence_length: int = 20
  sequence_period: Optional[int] = None
  learning_rate: float = 1e-3
  adam_momentum_decay: float = 0.0
  adam_variance_decay: float = 0.99
  discount: float = 0.99
  entropy_cost: float = 0.01
  baseline_cost: float = 0.5
  max_abs_reward: float = np.inf
  max_gradient_norm: float = np.inf

  # Replay options
  replay_table_name: str = adders_reverb.DEFAULT_PRIORITY_TABLE
  num_prefetch_threads: Optional[int] = None
  samples_per_insert: Optional[float] = None
  max_queue_size: Union[int, types.Batches] = 10_000

  def __post_init__(self):
    if isinstance(self.max_queue_size, types.Batches):
      self.max_queue_size *= self.batch_size
    assert self.max_queue_size > self.batch_size + 1, ("""
        max_queue_size must be strictly larger than the batch size:
        - during the last step in an episode we might write 2 sequences to
          Reverb at once (that's how SequenceAdder works)
        - Reverb does insertion/sampling in multiple threads, so data is
          added asynchronously at unpredictable times. Therefore we need
          additional buffer size in order to avoid deadlocks.""")



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the environment loop."""

from typing import Optional
from absl.testing import absltest
from absl.testing import parameterized

from acme import environment_loop
from acme import specs
from acme import types
from acme.testing import fakes
import numpy as np

EPISODE_LENGTH = 10

# Discount specs
F32_2_MIN_0_MAX_1 = specs.BoundedArray(
    dtype=np.float32, shape=(2,), minimum=0.0, maximum=1.0)
F32_2x1_MIN_0_MAX_1 = specs.BoundedArray(
    dtype=np.float32, shape=(2, 1), minimum=0.0, maximum=1.0)
TREE_MIN_0_MAX_1 = {'a': F32_2_MIN_0_MAX_1, 'b': F32_2x1_MIN_0_MAX_1}

# Reward specs
F32 = specs.Array(dtype=np.float32, shape=())
F32_1x3 = specs.Array(dtype=np.float32, shape=(1, 3))
TREE = {'a': F32, 'b': F32_1x3}

TEST_CASES = (
    ('scalar_discount_scalar_reward', None, None),
    ('vector_discount_scalar_reward', F32_2_MIN_0_MAX_1, F32),
    ('matrix_discount_matrix_reward', F32_2x1_MIN_0_MAX_1, F32_1x3),
    ('tree_discount_tree_reward', TREE_MIN_0_MAX_1, TREE),
    )


class EnvironmentLoopTest(parameterized.TestCase):

  @parameterized.named_parameters(*TEST_CASES)
  def test_one_episode(self, discount_spec, reward_spec):
    _, loop = _parameterized_setup(discount_spec, reward_spec)
    result = loop.run_episode()
    self.assertIn('episode_length', result)
    self.assertEqual(EPISODE_LENGTH, result['episode_length'])
    self.assertIn('episode_return', result)
    self.assertIn('steps_per_second', result)

  @parameterized.named_parameters(*TEST_CASES)
  def test_run_episodes(self, discount_spec, reward_spec):
    actor, loop = _parameterized_setup(discount_spec, reward_spec)

    # Run the loop. There should be EPISODE_LENGTH update calls per episode.
    loop.run(num_episodes=10)
    self.assertEqual(actor.num_updates, 10 * EPISODE_LENGTH)

  @parameterized.named_parameters(*TEST_CASES)
  def test_run_steps(self, discount_spec, reward_spec):
    actor, loop = _parameterized_setup(discount_spec, reward_spec)

    # Run the loop. This will run 2 episodes so that total number of steps is
    # at least 15.
    loop.run(num_steps=EPISODE_LENGTH + 5)
    self.assertEqual(actor.num_updates, 2 * EPISODE_LENGTH)


def _parameterized_setup(discount_spec: Optional[types.NestedSpec] = None,
                         reward_spec: Optional[types.NestedSpec] = None):
  """Common setup code that, unlike self.setUp, takes arguments.

  Args:
    discount_spec: None, or a (nested) specs.BoundedArray.
    reward_spec: None, or a (nested) specs.Array.
  Returns:
    environment, actor, loop
  """
  env_kwargs = {'episode_length': EPISODE_LENGTH}
  if discount_spec:
    env_kwargs['discount_spec'] = discount_spec
  if reward_spec:
    env_kwargs['reward_spec'] = reward_spec

  environment = fakes.DiscreteEnvironment(**env_kwargs)
  actor = fakes.Actor(specs.make_environment_spec(environment))
  loop = environment_loop.EnvironmentLoop(environment, actor)
  return actor, loop


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities related to loading TFDS datasets."""

import logging
from typing import Any, Dict, Iterator, Tuple

from acme import types
import jax
import jax.numpy as jnp
import tensorflow as tf
import tensorflow_datasets as tfds


def _episode_to_transition(step: Dict[str, Any]) -> types.Transition:
  return types.Transition(
      observation=step['observation'][:-1],
      action=step['action'][:-1],
      reward=step['reward'][:-1],
      discount=1.0 - tf.cast(step['is_terminal'][1:], dtype=tf.float32),
      # If next step is terminal, then the observation may be arbitrary.
      next_observation=step['observation'][1:],
  )


def _episode_steps_to_transition(episode) -> tf.data.Dataset:
  """Transforms an Episode into a dataset of Transitions."""
  episode = episode['steps']
  # The code below might fail if the dataset contains more than 1e9 transitions,
  # which is quite unlikely.
  data = tf.data.experimental.get_single_element(episode.batch(1000000000))
  data = _episode_to_transition(data)
  return tf.data.Dataset.from_tensor_slices(data)


def get_tfds_dataset(dataset_name: str):
  dataset = tfds.load(dataset_name)
  return dataset['train'].flat_map(_episode_steps_to_transition)


class JaxInMemoryRandomSampleIterator(Iterator[Any]):
  """In memory random sample iterator implemented in JAX.

  Loads the whole dataset in memory and performs random sampling with
  replacement of batches of `batch_size`.
  This class provides much faster sampling functionality compared to using
  an iterator on tf.data.Dataset.
  """

  def __init__(self,
               dataset: tf.data.Dataset,
               key: jnp.ndarray,
               batch_size: int):
    """Creates an iterator.

    Args:
      dataset: underlying tf Dataset
      key: a key to be used for random number generation
      batch_size: batch size
    """
    # Read the whole dataset. We use artificially large batch_size to make sure
    # we capture the whole dataset.
    data = next(dataset.batch(1000000000).as_numpy_iterator())
    self._dataset_size = jax.tree_flatten(
        jax.tree_map(lambda x: x.shape[0], data))[0][0]
    self._jax_dataset = jax.tree_map(jnp.asarray, data)
    logging.info('Finished loading a dataset into memory. Elements: %d',
                 self._dataset_size)
    self._key = key

    def sample(key: jnp.ndarray) -> Tuple[Any, jnp.ndarray]:
      key, key_randint = jax.random.split(key)
      indices = jax.random.randint(key_randint, (batch_size,), minval=0,
                                   maxval=self._dataset_size)
      demo_transitions = jax.tree_map(lambda d: jnp.take(d, indices, axis=0),
                                      self._jax_dataset)
      return demo_transitions, key
    self._sample = jax.jit(sample)

  def __next__(self) -> Any:
    data, self._key = self._sample(self._key)
    return data

  @property
  def dataset_size(self):
    """An integer of the dataset cardinality."""
    return self._dataset_size



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Recurrent DQfD (R2D3) agent implementation."""

import functools
from typing import Optional

from acme import datasets
from acme import specs
from acme import types as acme_types
from acme.adders import reverb as adders
from acme.agents import agent
from acme.agents.tf import actors
from acme.agents.tf.r2d2 import learning
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils
from acme.utils import counting
from acme.utils import loggers
import reverb
import sonnet as snt
import tensorflow as tf
import tree
import trfl


class R2D3(agent.Agent):
  """R2D3 Agent.

  This implements a single-process R2D2 agent that mixes demonstrations with
  actor experience.
  """

  def __init__(self,
               environment_spec: specs.EnvironmentSpec,
               network: snt.RNNCore,
               target_network: snt.RNNCore,
               burn_in_length: int,
               trace_length: int,
               replay_period: int,
               demonstration_dataset: tf.data.Dataset,
               demonstration_ratio: float,
               counter: Optional[counting.Counter] = None,
               logger: Optional[loggers.Logger] = None,
               discount: float = 0.99,
               batch_size: int = 32,
               target_update_period: int = 100,
               importance_sampling_exponent: float = 0.2,
               epsilon: float = 0.01,
               learning_rate: float = 1e-3,
               log_to_bigtable: bool = False,
               log_name: str = 'agent',
               checkpoint: bool = True,
               min_replay_size: int = 1000,
               max_replay_size: int = 1000000,
               samples_per_insert: float = 32.0):

    sequence_length = burn_in_length + trace_length + 1
    extra_spec = {
        'core_state': network.initial_state(1),
    }
    # Remove batch dimensions.
    extra_spec = tf2_utils.squeeze_batch_dim(extra_spec)
    replay_table = reverb.Table(
        name=adders.DEFAULT_PRIORITY_TABLE,
        sampler=reverb.selectors.Uniform(),
        remover=reverb.selectors.Fifo(),
        max_size=max_replay_size,
        rate_limiter=reverb.rate_limiters.MinSize(min_size_to_sample=1),
        signature=adders.SequenceAdder.signature(
            environment_spec, extra_spec, sequence_length=sequence_length))
    self._server = reverb.Server([replay_table], port=None)
    address = f'localhost:{self._server.port}'

    # Component to add things into replay.
    sequence_kwargs = dict(
        period=replay_period,
        sequence_length=sequence_length,
    )
    adder = adders.SequenceAdder(client=reverb.Client(address),
                                   **sequence_kwargs)

    # The dataset object to learn from.
    dataset = datasets.make_reverb_dataset(
        server_address=address)

    # Combine with demonstration dataset.
    transition = functools.partial(_sequence_from_episode,
                                   extra_spec=extra_spec,
                                   **sequence_kwargs)
    dataset_demos = demonstration_dataset.map(transition)
    dataset = tf.data.experimental.sample_from_datasets(
        [dataset, dataset_demos],
        [1 - demonstration_ratio, demonstration_ratio])

    # Batch and prefetch.
    dataset = dataset.batch(batch_size, drop_remainder=True)
    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)

    tf2_utils.create_variables(network, [environment_spec.observations])
    tf2_utils.create_variables(target_network, [environment_spec.observations])

    learner = learning.R2D2Learner(
        environment_spec=environment_spec,
        network=network,
        target_network=target_network,
        burn_in_length=burn_in_length,
        dataset=dataset,
        reverb_client=reverb.TFClient(address),
        counter=counter,
        logger=logger,
        sequence_length=sequence_length,
        discount=discount,
        target_update_period=target_update_period,
        importance_sampling_exponent=importance_sampling_exponent,
        max_replay_size=max_replay_size,
        learning_rate=learning_rate,
        store_lstm_state=False,
    )

    self._checkpointer = tf2_savers.Checkpointer(
        subdirectory='r2d2_learner',
        time_delta_minutes=60,
        objects_to_save=learner.state,
        enable_checkpointing=checkpoint,
    )

    self._snapshotter = tf2_savers.Snapshotter(
        objects_to_save={'network': network}, time_delta_minutes=60.)

    policy_network = snt.DeepRNN([
        network,
        lambda qs: trfl.epsilon_greedy(qs, epsilon=epsilon).sample(),
    ])

    actor = actors.RecurrentActor(policy_network, adder)
    observations_per_step = (float(replay_period * batch_size) /
                             samples_per_insert)
    super().__init__(
        actor=actor,
        learner=learner,
        min_observations=replay_period * max(batch_size, min_replay_size),
        observations_per_step=observations_per_step)

  def update(self):
    super().update()
    self._snapshotter.save()
    self._checkpointer.save()


def _sequence_from_episode(observations: acme_types.NestedTensor,
                           actions: tf.Tensor,
                           rewards: tf.Tensor,
                           discounts: tf.Tensor,
                           extra_spec: acme_types.NestedSpec,
                           period: int,
                           sequence_length: int):
  """Produce Reverb-like sequence from a full episode.

  Observations, actions, rewards and discounts have the same length. This
  function will ignore the first reward and discount and the last action.

  This function generates fake (all-zero) extras.

  See docs for reverb.SequenceAdder() for more details.

  Args:
    observations: [L, ...] Tensor.
    actions: [L, ...] Tensor.
    rewards: [L] Tensor.
    discounts: [L] Tensor.
    extra_spec: A possibly nested structure of specs for extras. This function
      will generate fake (all-zero) extras.
    period: The period with which we add sequences.
    sequence_length: The fixed length of sequences we wish to add.

  Returns:
    (o_t, a_t, r_t, d_t, e_t) Tuple.
  """

  length = tf.shape(rewards)[0]
  first = tf.random.uniform(shape=(), minval=0, maxval=length, dtype=tf.int32)
  first = first // period * period  # Get a multiple of `period`.
  to = tf.minimum(first + sequence_length, length)

  def _slice_and_pad(x):
    pad_length = sequence_length + first - to
    padding_shape = tf.concat([[pad_length], tf.shape(x)[1:]], axis=0)
    result = tf.concat([x[first:to], tf.zeros(padding_shape, x.dtype)], axis=0)
    result.set_shape([sequence_length] + x.shape.as_list()[1:])
    return result

  o_t = tree.map_structure(_slice_and_pad, observations)
  a_t = tree.map_structure(_slice_and_pad, actions)
  r_t = _slice_and_pad(rewards)
  d_t = _slice_and_pad(discounts)
  start_of_episode = tf.equal(first, 0)
  start_of_episode = tf.expand_dims(start_of_episode, axis=0)
  start_of_episode = tf.tile(start_of_episode, [sequence_length])

  def _sequence_zeros(spec):
    return tf.zeros([sequence_length] + spec.shape, spec.dtype)

  e_t = tree.map_structure(_sequence_zeros, extra_spec)

  key = tf.zeros([], tf.uint64)
  probability = tf.ones([], tf.float64)
  table_size = tf.ones([], tf.int64)
  priority = tf.ones([], tf.float64)
  info = reverb.SampleInfo(
      key=key,
      probability=probability,
      table_size=table_size,
      priority=priority)
  return reverb.ReplaySample(
      info=info,
      data=adders.Step(
          observation=o_t,
          action=a_t,
          reward=r_t,
          discount=d_t,
          start_of_episode=start_of_episode,
          extras=e_t))



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the SVG agent."""

import sys
from typing import Dict, Sequence

from absl.testing import absltest
import acme
from acme import specs
from acme import types
from acme.agents.tf import svg0_prior
from acme.testing import fakes
from acme.tf import networks
from acme.tf import utils as tf2_utils

import numpy as np
import sonnet as snt


def make_networks(
    action_spec: types.NestedSpec,
    policy_layer_sizes: Sequence[int] = (10, 10),
    critic_layer_sizes: Sequence[int] = (10, 10),
) -> Dict[str, snt.Module]:
  """Creates networks used by the agent."""
  # Get total number of action dimensions from action spec.
  num_dimensions = np.prod(action_spec.shape, dtype=int)

  policy_network = snt.Sequential([
      tf2_utils.batch_concat,
      networks.LayerNormMLP(policy_layer_sizes, activate_final=True),
      networks.MultivariateNormalDiagHead(
          num_dimensions,
          tanh_mean=True,
          min_scale=0.3,
          init_scale=0.7,
          fixed_scale=False,
          use_tfd_independent=False)
  ])
  # The multiplexer concatenates the (maybe transformed) observations/actions.
  multiplexer = networks.CriticMultiplexer()
  critic_network = snt.Sequential([
      multiplexer,
      networks.LayerNormMLP(critic_layer_sizes, activate_final=True),
      networks.NearZeroInitializedLinear(1),
  ])

  return {
      'policy': policy_network,
      'critic': critic_network,
  }


class SVG0Test(absltest.TestCase):

  def test_svg0(self):
    # Create a fake environment to test with.
    environment = fakes.ContinuousEnvironment(episode_length=10)
    spec = specs.make_environment_spec(environment)

    # Create the networks.
    agent_networks = make_networks(spec.actions)

    # Construct the agent.
    agent = svg0_prior.SVG0(
        environment_spec=spec,
        policy_network=agent_networks['policy'],
        critic_network=agent_networks['critic'],
        batch_size=10,
        samples_per_insert=2,
        min_replay_size=10,
    )

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent)
    loop.run(num_episodes=2)
    # Make sure Acme doesn't import excessive number of modules.
    self.assertLess(len(sys.modules), 4500)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of a behavior cloning (BC) agent."""

from acme.agents.jax.bc import pretraining
from acme.agents.jax.bc.learning import BCLearner
from acme.agents.jax.bc.losses import logp
from acme.agents.jax.bc.losses import Loss
from acme.agents.jax.bc.losses import mse
from acme.agents.jax.bc.losses import peerbc
from acme.agents.jax.bc.losses import rcal



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logger for writing to an in-memory list.

This is convenient for e.g. interactive usage via Google Colab.

For example, for usage with pandas:

```python
from acme.utils import loggers
import pandas as pd

logger = InMemoryLogger()
# ...
logger.write({'foo': 1.337, 'bar': 420})

results = pd.DataFrame(logger.data)
```
"""

from typing import Sequence

from acme.utils.loggers import base


class InMemoryLogger(base.Logger):
  """A simple logger that keeps all data in memory."""

  def __init__(self):
    self._data = []

  def write(self, data: base.LoggingData):
    self._data.append(data)

  def close(self):
    pass

  @property
  def data(self) -> Sequence[base.LoggingData]:
    return self._data



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for acme.tf.losses.distributional."""

from absl.testing import absltest
from absl.testing import parameterized
from acme.tf.losses import distributional
import numpy as np
from numpy import testing as npt
import tensorflow as tf


def _reference_l2_project(src_support, src_probs, dst_support):
  """Multi-axis l2_project, implemented using single-axis l2_project.

  This is for testing multiaxis_l2_project's consistency with l2_project,
  when used with multi-axis support vs single-axis support.

  Args:
    src_support: Zp in l2_project.
    src_probs: P in l2_project.
    dst_support: Zq in l2_project.

  Returns:
    src_probs, projected onto dst_support.
  """
  assert src_support.shape == src_probs.shape

  # Remove the batch and value axes, and broadcast the rest to a common shape.
  common_shape = np.broadcast(src_support[0, ..., 0],
                              dst_support[..., 0]).shape

  # If src_* have fewer internal axes than len(common_shape), insert size-1
  # axes.
  while src_support.ndim-2 < len(common_shape):
    src_support = src_support[:, None, ...]

  src_probs = np.reshape(src_probs, src_support.shape)

  # Broadcast args' non-batch, non-value axes to common_shape.
  src_support = np.broadcast_to(
      src_support,
      src_support.shape[:1] + common_shape + src_support.shape[-1:])
  src_probs = np.broadcast_to(src_probs, src_support.shape)
  dst_support = np.broadcast_to(
      dst_support,
      common_shape + dst_support.shape[-1:])

  output_shape = (src_support.shape[0],) + dst_support.shape

  # Collapse all but the first (batch) and last (atom) axes.
  src_support = src_support.reshape(
      [src_support.shape[0], -1, src_support.shape[-1]])
  src_probs = src_probs.reshape(
      [src_probs.shape[0], -1, src_probs.shape[-1]])

  # Collapse all but the last (atom) axes.
  dst_support = dst_support.reshape([-1, dst_support.shape[-1]])

  dst_probs = np.zeros(src_support.shape[:1] + dst_support.shape,
                       dtype=src_probs.dtype)

  # iterate over all supports
  for i in range(src_support.shape[1]):
    s_support = tf.convert_to_tensor(src_support[:, i, :])
    s_probs = tf.convert_to_tensor(src_probs[:, i, :])
    d_support = tf.convert_to_tensor(dst_support[i, :])
    d_probs = distributional.l2_project(s_support, s_probs, d_support)
    dst_probs[:, i, :] = d_probs.numpy()

  return dst_probs.reshape(output_shape)


class L2ProjectTest(parameterized.TestCase):

  @parameterized.parameters(
      [(2, 11), (11,)],  # C = (), D = (), matching num_atoms (11 and 11)
      [(2, 11), (5,)],  # C = (), D = (), differing num_atoms (11 and 5).
      [(2, 3, 11), (3, 5)],  # C = (3,), D = (3,)
      [(2, 1, 11), (3, 5)],  # C = (1,), D = (3,)
      [(2, 3, 11), (1, 5)],  # (C = (3,), D = (1,)
      [(2, 3, 4, 11), (3, 4, 5)],  # C = (3, 4), D = (3, 4)
      [(2, 3, 4, 11), (4, 5)],  # C = (3, 4), D = (4,)
      [(2, 4, 11), (3, 4, 5)],  # C = (4,), D = (3, 4)
  )
  def test_multiaxis(self, src_shape, dst_shape):
    """Tests consistency between multi-axis and single-axis l2_project.

    This calls l2_project on multi-axis supports, and checks that it gets
    the same outcomes as many calls to single-axis supports.

    Args:
      src_shape: Shape of source support. Includes a leading batch axis.
      dst_shape: Shape of destination support.
        Does not include a leading batch axis.
    """
    # src_shape includes a leading batch axis, whereas dst_shape does not.
    # assert len(src_shape) >= (1 + len(dst_shape))

    def make_support(shape, minimum):
      """Creates a ndarray of supports."""
      values = np.linspace(start=minimum, stop=minimum+100, num=shape[-1])
      offsets = np.arange(np.prod(shape[:-1]))
      result = values[None, :] + offsets[:, None]
      return result.reshape(shape)

    src_support = make_support(src_shape, -1)
    dst_support = make_support(dst_shape, -.75)

    rng = np.random.RandomState(1)
    src_probs = rng.uniform(low=1.0, high=2.0, size=src_shape)
    src_probs /= src_probs.sum()

    # Repeated calls to l2_project using single-axis supports.
    expected_dst_probs = _reference_l2_project(src_support,
                                               src_probs,
                                               dst_support)

    # A single call to l2_project, with multi-axis supports.
    dst_probs = distributional.multiaxis_l2_project(
        tf.convert_to_tensor(src_support),
        tf.convert_to_tensor(src_probs),
        tf.convert_to_tensor(dst_support)).numpy()

    npt.assert_allclose(dst_probs, expected_dst_probs)

  @parameterized.parameters(
      # Same src and dst support shape, dst support is shifted by +.25
      ([[0., 1, 2, 3]],
       [[0., 1, 0, 0]],
       [.25, 1.25, 2.25, 3.25],
       [[.25, .75, 0, 0]]),
      # Similar to above, but with batched src.
      ([[0., 1, 2, 3],
        [0., 1, 2, 3]],
       [[0., 1, 0, 0],
        [0., 0, 1, 0]],
       [.25, 1.25, 2.25, 3.25],
       [[.25, .75, 0, 0],
        [0., .25, .75, 0]]),
      # Similar to above, but src_probs has two 0.5's, instead of being one-hot.
      ([[0., 1, 2, 3]],
       [[0., .5, .5, 0]],
       [.25, 1.25, 2.25, 3.25],
       0.5 * (np.array([[.25, .75, 0, 0]]) + np.array([[0., .25, .75, 0]]))),
      # src and dst support have differing sizes
      ([[0., 1, 2, 3]],
       [[0., 1, 0, 0]],
       [0.00, 0.25, 0.50, 0.75, 1.00, 1.25, 1.50, 1.75, 2.00, 2.25, 2.50],
       [[0.00, 0.00, 0.00, 0.00, 1.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00]]),
      )
  def test_l2_projection(
      self, src_support, src_probs, dst_support, expected_dst_probs):

    dst_probs = distributional.multiaxis_l2_project(
        tf.convert_to_tensor(src_support),
        tf.convert_to_tensor(src_probs),
        tf.convert_to_tensor(dst_support)).numpy()
    npt.assert_allclose(dst_probs, expected_dst_probs)


if __name__ == '__main__':
  absltest.main()




# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ActorCore interface definition."""

import dataclasses
from typing import Callable, Generic, Mapping, Tuple, TypeVar, Union

from acme import types
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.jax.types import PRNGKey
import chex
import jax
import jax.numpy as jnp


NoneType = type(None)


# The state of the actor. This could include recurrent network state or any
# other state which needs to be propagated through the select_action calls.
State = TypeVar('State')
# The extras to be passed to the observe method.
Extras = TypeVar('Extras')
RecurrentState = TypeVar('RecurrentState')


@dataclasses.dataclass
class ActorCore(Generic[State, Extras]):
  """Pure functions that define the algorithm-specific actor functionality."""
  init: Callable[[PRNGKey], State]
  select_action: Callable[[
      networks_lib.Params, networks_lib.Observation, State
  ], Tuple[networks_lib.Action, State]]
  get_extras: Callable[[State], Extras]


# A simple feed forward policy which produces no extras and takes only an RNGKey
# as a state.
FeedForwardPolicy = Callable[
    [networks_lib.Params, PRNGKey, networks_lib.Observation],
    networks_lib.Action]

FeedForwardPolicyWithExtra = Callable[
    [networks_lib.Params, PRNGKey, networks_lib.Observation],
    Tuple[networks_lib.Action, types.NestedArray]]

RecurrentPolicy = Callable[[
    networks_lib.Params, PRNGKey, networks_lib
    .Observation, RecurrentState
], Tuple[networks_lib.Action, RecurrentState]]

Policy = Union[FeedForwardPolicy, FeedForwardPolicyWithExtra, RecurrentPolicy]


def batched_feed_forward_to_actor_core(
    policy: FeedForwardPolicy
) -> ActorCore[PRNGKey, NoneType]:
  """A convenience adaptor from FeedForwardPolicy to ActorCore."""

  def select_action(params: networks_lib.Params,
                    observation: networks_lib.Observation,
                    state: PRNGKey):
    rng = state
    rng1, rng2 = jax.random.split(rng)
    observation = utils.add_batch_dim(observation)
    action = utils.squeeze_batch_dim(policy(params, rng1, observation))
    return action, rng2

  def init(rng: PRNGKey) -> PRNGKey:
    return rng

  def get_extras(unused_rng: PRNGKey) -> NoneType:
    return None
  return ActorCore(init=init, select_action=select_action,
                   get_extras=get_extras)


@chex.dataclass(frozen=True, mappable_dataclass=False)
class SimpleActorCoreStateWithExtras:
  rng: PRNGKey
  extras: Mapping[str, jnp.ndarray]


def batched_feed_forward_with_extras_to_actor_core(
    policy: FeedForwardPolicyWithExtra
) -> ActorCore[SimpleActorCoreStateWithExtras, Mapping[str, jnp.ndarray]]:
  """A convenience adaptor from FeedForwardPolicy to ActorCore."""

  def select_action(params: networks_lib.Params,
                    observation: networks_lib.Observation,
                    state: SimpleActorCoreStateWithExtras):
    rng = state.rng
    rng1, rng2 = jax.random.split(rng)
    observation = utils.add_batch_dim(observation)
    action, extras = utils.squeeze_batch_dim(policy(params, rng1, observation))
    return action, SimpleActorCoreStateWithExtras(rng2, extras)

  def init(rng: PRNGKey) -> SimpleActorCoreStateWithExtras:
    return SimpleActorCoreStateWithExtras(rng, {})

  def get_extras(
      state: SimpleActorCoreStateWithExtras) -> Mapping[str, jnp.ndarray]:
    return state.extras
  return ActorCore(init=init, select_action=select_action,
                   get_extras=get_extras)


@chex.dataclass(frozen=True, mappable_dataclass=False)
class SimpleActorCoreRecurrentState(Generic[RecurrentState]):
  rng: PRNGKey
  recurrent_state: RecurrentState


def batched_recurrent_to_actor_core(
    recurrent_policy: RecurrentPolicy, initial_core_state: RecurrentState
) -> ActorCore[SimpleActorCoreRecurrentState[RecurrentState], Mapping[
    str, jnp.ndarray]]:
  """Returns ActorCore for a recurrent policy."""
  def select_action(params: networks_lib.Params,
                    observation: networks_lib.Observation,
                    state: SimpleActorCoreRecurrentState[RecurrentState]):
    # TODO(b/161332815): Make JAX Actor work with batched or unbatched inputs.
    rng = state.rng
    rng, policy_rng = jax.random.split(rng)
    observation = utils.add_batch_dim(observation)
    recurrent_state = utils.add_batch_dim(state.recurrent_state)
    action, new_recurrent_state = utils.squeeze_batch_dim(recurrent_policy(
        params, policy_rng, observation, recurrent_state))
    return action, SimpleActorCoreRecurrentState(rng, new_recurrent_state)

  initial_core_state = utils.squeeze_batch_dim(initial_core_state)
  def init(rng: PRNGKey) -> SimpleActorCoreRecurrentState[RecurrentState]:
    return SimpleActorCoreRecurrentState(rng, initial_core_state)

  def get_extras(
      state: SimpleActorCoreRecurrentState[RecurrentState]
  ) -> Mapping[str, jnp.ndarray]:
    return {'core_state': state.recurrent_state}

  return ActorCore(init=init, select_action=select_action,
                   get_extras=get_extras)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared helpers for different experiment flavours."""

import functools
from typing import Mapping, Sequence, Optional

from acme import specs
from acme import types
from acme.agents.tf.svg0_prior import utils as svg0_utils
from acme.tf import networks
from acme.tf import utils as tf2_utils

import numpy as np
import sonnet as snt


def make_default_networks(
    action_spec: specs.BoundedArray,
    policy_layer_sizes: Sequence[int] = (256, 256, 256),
    critic_layer_sizes: Sequence[int] = (512, 512, 256),
) -> Mapping[str, types.TensorTransformation]:
  """Creates networks used by the agent."""

  # Get total number of action dimensions from action spec.
  num_dimensions = np.prod(action_spec.shape, dtype=int)

  policy_network = snt.Sequential([
      tf2_utils.batch_concat,
      networks.LayerNormMLP(policy_layer_sizes, activate_final=True),
      networks.MultivariateNormalDiagHead(
          num_dimensions,
          tanh_mean=True,
          min_scale=0.3,
          init_scale=0.7,
          fixed_scale=False,
          use_tfd_independent=False)
  ])
  # The multiplexer concatenates the (maybe transformed) observations/actions.
  multiplexer = networks.CriticMultiplexer(
      action_network=networks.ClipToSpec(action_spec))
  critic_network = snt.Sequential([
      multiplexer,
      networks.LayerNormMLP(critic_layer_sizes, activate_final=True),
      networks.NearZeroInitializedLinear(1),
  ])

  return {
      "policy": policy_network,
      "critic": critic_network,
  }


def make_network_with_prior(
    action_spec: specs.BoundedArray,
    policy_layer_sizes: Sequence[int] = (200, 100),
    critic_layer_sizes: Sequence[int] = (400, 300),
    prior_layer_sizes: Sequence[int] = (200, 100),
    policy_keys: Optional[Sequence[str]] = None,
    prior_keys: Optional[Sequence[str]] = None,
) -> Mapping[str, types.TensorTransformation]:
  """Creates networks used by the agent."""

  # Get total number of action dimensions from action spec.
  num_dimensions = np.prod(action_spec.shape, dtype=int)
  flatten_concat_policy = functools.partial(
      svg0_utils.batch_concat_selection, concat_keys=policy_keys)
  flatten_concat_prior = functools.partial(
      svg0_utils.batch_concat_selection, concat_keys=prior_keys)

  policy_network = snt.Sequential([
      flatten_concat_policy,
      networks.LayerNormMLP(policy_layer_sizes, activate_final=True),
      networks.MultivariateNormalDiagHead(
          num_dimensions,
          tanh_mean=True,
          min_scale=0.1,
          init_scale=0.7,
          fixed_scale=False,
          use_tfd_independent=False)
  ])
  # The multiplexer concatenates the (maybe transformed) observations/actions.
  multiplexer = networks.CriticMultiplexer(
      observation_network=flatten_concat_policy,
      action_network=networks.ClipToSpec(action_spec))
  critic_network = snt.Sequential([
      multiplexer,
      networks.LayerNormMLP(critic_layer_sizes, activate_final=True),
      networks.NearZeroInitializedLinear(1),
  ])
  prior_network = snt.Sequential([
      flatten_concat_prior,
      networks.LayerNormMLP(prior_layer_sizes, activate_final=True),
      networks.MultivariateNormalDiagHead(
          num_dimensions,
          tanh_mean=True,
          min_scale=0.1,
          init_scale=0.7,
          fixed_scale=False,
          use_tfd_independent=False)
  ])
  return {
      "policy": policy_network,
      "critic": critic_network,
      "prior": prior_network,
  }



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""D4PG Builder."""
import dataclasses
from typing import Callable, Iterator, List, Optional

import acme
from acme import adders
from acme import core
from acme import specs
from acme import types
from acme.adders import reverb as adders_reverb
from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax import actors
from acme.agents.jax import builders
from acme.agents.jax.d4pg import learning
from acme.datasets import reverb as datasets
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import optax
import reverb
from reverb import rate_limiters
import rlax


@dataclasses.dataclass
class D4PGConfig:
  """Configuration options for D4PG."""
  sigma: float = 0.3
  target_update_period: int = 100
  samples_per_insert: float = 32.0

  # Loss options
  n_step: int = 5
  discount: float = 0.99
  batch_size: int = 256
  learning_rate: float = 1e-4
  clipping: bool = True

  # Replay options
  min_replay_size: int = 1000
  max_replay_size: int = 1000000
  replay_table_name: str = adders_reverb.DEFAULT_PRIORITY_TABLE
  prefetch_size: int = 4
  # Rate to be used for the SampleToInsertRatio rate limitter tolerance.
  # See a formula in make_replay_tables for more details.
  samples_per_insert_tolerance_rate: float = 0.1

  # How many gradient updates to perform per step.
  num_sgd_steps_per_step: int = 1


@dataclasses.dataclass
class D4PGNetworks:
  """Network and pure functions for the D4PG agent.."""
  policy_network: networks_lib.FeedForwardNetwork
  critic_network: networks_lib.FeedForwardNetwork


def get_default_behavior_policy(networks: D4PGNetworks, config: D4PGConfig):
  # Selects action according to the policy
  def behavior_policy(params: networks_lib.Params, key: networks_lib.PRNGKey,
                      observation: types.NestedArray):
    action = networks.policy_network.apply(params, observation)
    noisy_action = rlax.add_gaussian_noise(key, action, config.sigma)
    return noisy_action
  return behavior_policy


def get_default_eval_policy(networks: D4PGNetworks):
  # Selects action according to the policy
  def behavior_policy(params: networks_lib.Params, key: networks_lib.PRNGKey,
                      observation: types.NestedArray):
    del key  # Evaluation policy is deterministic.
    action = networks.policy_network.apply(params, observation)
    return action
  return behavior_policy


class D4PGBuilder(builders.ActorLearnerBuilder):
  """D4PG Builder."""

  def __init__(
      self,
      config: D4PGConfig,
      logger_fn: Callable[[], loggers.Logger] = lambda: None,
  ):
    """Creates a D4PG learner, a behavior policy and an eval actor.

    Args:
      config: a config with D4PG hps
      logger_fn: a logger factory for the learner
    """
    self._config = config
    self._logger_fn = logger_fn

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: D4PGNetworks,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:
    policy_optimizer = optax.adam(self._config.learning_rate)
    critic_optimizer = optax.adam(self._config.learning_rate)

    if self._config.clipping:
      policy_optimizer = optax.chain(
          optax.clip_by_global_norm(40.), policy_optimizer)
      critic_optimizer = optax.chain(
          optax.clip_by_global_norm(40.), critic_optimizer)

    # The learner updates the parameters (and initializes them).
    return learning.D4PGLearner(
        policy_network=networks.policy_network,
        critic_network=networks.critic_network,
        random_key=random_key,
        policy_optimizer=policy_optimizer,
        critic_optimizer=critic_optimizer,
        clipping=self._config.clipping,
        discount=self._config.discount,
        target_update_period=self._config.target_update_period,
        iterator=dataset,
        counter=counter,
        logger=self._logger_fn(),
        num_sgd_steps_per_step=self._config.num_sgd_steps_per_step)

  def make_replay_tables(
      self,
      environment_spec: specs.EnvironmentSpec,
  ) -> List[reverb.Table]:
    """Create tables to insert data into."""
    samples_per_insert_tolerance = (
        self._config.samples_per_insert_tolerance_rate *
        self._config.samples_per_insert)
    error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
    limiter = rate_limiters.SampleToInsertRatio(
        min_size_to_sample=self._config.min_replay_size,
        samples_per_insert=self._config.samples_per_insert,
        error_buffer=error_buffer)
    return [
        reverb.Table(
            name=self._config.replay_table_name,
            sampler=reverb.selectors.Uniform(),
            remover=reverb.selectors.Fifo(),
            max_size=self._config.max_replay_size,
            rate_limiter=limiter,
            signature=adders_reverb.NStepTransitionAdder.signature(
                environment_spec))
    ]

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    """Create a dataset iterator to use for learning/updating the agent."""
    dataset = datasets.make_reverb_dataset(
        table=self._config.replay_table_name,
        server_address=replay_client.server_address,
        batch_size=self._config.batch_size *
        self._config.num_sgd_steps_per_step,
        prefetch_size=self._config.prefetch_size)
    return dataset.as_numpy_iterator()

  def make_adder(self,
                 replay_client: reverb.Client) -> Optional[adders.Adder]:
    """Create an adder which records data generated by the actor/environment."""
    return adders_reverb.NStepTransitionAdder(
        priority_fns={self._config.replay_table_name: None},
        client=replay_client,
        n_step=self._config.n_step,
        discount=self._config.discount)

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None) -> acme.Actor:
    assert variable_source is not None
    actor_core = actor_core_lib.batched_feed_forward_to_actor_core(
        policy_network)
    # Inference happens on CPU, so it's better to move variables there too.
    variable_client = variable_utils.VariableClient(variable_source, 'policy',
                                                    device='cpu')
    return actors.GenericActor(
        actor_core, random_key, variable_client, adder, backend='cpu')



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PPO config."""
import dataclasses

from acme.adders import reverb as adders_reverb
import rlax


@dataclasses.dataclass
class R2D2Config:
  """Configuration options for R2D2 agent."""
  discount: float = 0.997
  target_update_period: int = 2500
  evaluation_epsilon: float = 0.
  num_epsilons: int = 256
  variable_update_period: int = 400

  # Learner options
  burn_in_length: int = 40
  trace_length: int = 80
  sequence_period: int = 40
  learning_rate: float = 1e-3
  bootstrap_n: int = 5
  clip_rewards: bool = False
  tx_pair: rlax.TxPair = rlax.SIGNED_HYPERBOLIC_PAIR

  # Replay options
  samples_per_insert_tolerance_rate: float = 0.1
  samples_per_insert: float = 4.0
  min_replay_size: int = 50_000
  max_replay_size: int = 100_000
  batch_size: int = 64
  prefetch_size: int = 2
  num_parallel_calls: int = 16
  replay_table_name: str = adders_reverb.DEFAULT_PRIORITY_TABLE

  # Priority options
  importance_sampling_exponent: float = 0.6
  priority_exponent: float = 0.9
  max_priority_weight: float = 0.9



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for Reverb-based adders."""

from typing import Dict, Union

from acme import types
from acme.adders.reverb import base
import jax.numpy as jnp
import numpy as np
import tree


def zeros_like(x: Union[np.ndarray, int, float, np.number]):
  """Returns a zero-filled object of the same (d)type and shape as the input.

  The difference between this and `np.zeros_like()` is that this works well
  with `np.number`, `int`, `float`, and `jax.numpy.DeviceArray` objects without
  converting them to `np.ndarray`s.

  Args:
    x: The object to replace with 0s.

  Returns:
    A zero-filed object of the same (d)type and shape as the input.
  """
  if isinstance(x, (int, float, np.number)):
    return type(x)(0)
  elif isinstance(x, jnp.DeviceArray):
    return jnp.zeros_like(x)
  elif isinstance(x, np.ndarray):
    return np.zeros_like(x)
  else:
    raise ValueError(
        f'Input ({type(x)}) must be either a numpy array, an int, or a float.')


def final_step_like(step: base.Step,
                    next_observation: types.NestedArray) -> base.Step:
  """Return a list of steps with the final step zero-filled."""
  # Make zero-filled components so we can fill out the last step.
  zero_action, zero_reward, zero_discount, zero_extras = tree.map_structure(
      zeros_like, (step.action, step.reward, step.discount, step.extras))

  # Return a final step that only has next_observation.
  return base.Step(
      observation=next_observation,
      action=zero_action,
      reward=zero_reward,
      discount=zero_discount,
      start_of_episode=False,
      extras=zero_extras)


def calculate_priorities(
    priority_fns: base.PriorityFnMapping,
    trajectory_or_transition: Union[base.Trajectory, types.Transition],
) -> Dict[str, float]:
  """Helper used to calculate the priority of a Trajectory or Transition.

  This helper converts the leaves of the Trajectory or Transition from
  `reverb.TrajectoryColumn` objects into numpy arrays. The converted Trajectory
  or Transition is then passed into each of the functions in `priority_fns`.

  Args:
    priority_fns: a mapping from table names to priority functions (i.e. a
      callable of type PriorityFn). The given function will be used to generate
      the priority (a float) for the given table.
    trajectory_or_transition: the trajectory or transition used to compute
      priorities.

  Returns:
    A dictionary mapping from table names to the priority (a float) for the
    given collection Trajectory or Transition.
  """
  if any([priority_fn is not None for priority_fn in priority_fns.values()]):

    trajectory_or_transition = tree.map_structure(lambda col: col.numpy(),
                                                  trajectory_or_transition)

  return {
      table: (priority_fn(trajectory_or_transition) if priority_fn else 1.0)
      for table, priority_fn in priority_fns.items()
  }



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AIL learner implementation."""
import functools
import itertools
import time
from typing import Any, Callable, Iterator, List, NamedTuple, Optional, Tuple

import acme
from acme import types
from acme.agents.jax.ail import losses
from acme.agents.jax.ail import networks as ail_networks
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
from acme.utils import reverb_utils
import jax
import optax
import reverb


class DiscriminatorTrainingState(NamedTuple):
  """Contains training state for the discriminator."""
  # State of the optimizer used to optimize the discriminator parameters.
  optimizer_state: optax.OptState

  # Parameters of the discriminator.
  discriminator_params: networks_lib.Params

  # State of the discriminator
  discriminator_state: losses.State

  # For AIRL variants, we need the policy params to compute the loss.
  policy_params: Optional[networks_lib.Params]

  # Key for random number generation.
  key: networks_lib.PRNGKey

  # Training step of the discriminator.
  steps: int


class TrainingState(NamedTuple):
  """Contains training state of the AIL learner."""
  rewarder_state: DiscriminatorTrainingState
  learner_state: Any


def ail_update_step(
    state: DiscriminatorTrainingState, data: Tuple[types.Transition,
                                                   types.Transition],
    optimizer: optax.GradientTransformation,
    ail_network: ail_networks.AILNetworks,
    loss_fn: losses.Loss) -> Tuple[DiscriminatorTrainingState, losses.Metrics]:
  """Run an update steps on the given transitions.

  Args:
    state: The learner state.
    data: Demo and rb transitions.
    optimizer: Discriminator optimizer.
    ail_network: AIL networks.
    loss_fn: Discriminator loss to minimize.

  Returns:
    A new state and metrics.
  """
  demo_transitions, rb_transitions = data
  key, discriminator_key, loss_key = jax.random.split(state.key, 3)

  def compute_loss(
      discriminator_params: networks_lib.Params) -> losses.LossOutput:
    discriminator_fn = functools.partial(
        ail_network.discriminator_network.apply,
        discriminator_params,
        state.policy_params,
        is_training=True,
        rng=discriminator_key)
    return loss_fn(discriminator_fn, state.discriminator_state,
                   demo_transitions, rb_transitions, loss_key)

  loss_grad = jax.grad(compute_loss, has_aux=True)

  grads, (loss, new_discriminator_state) = loss_grad(state.discriminator_params)

  update, optimizer_state = optimizer.update(
      grads,
      state.optimizer_state,
      params=state.discriminator_params)
  discriminator_params = optax.apply_updates(state.discriminator_params, update)

  new_state = DiscriminatorTrainingState(
      optimizer_state=optimizer_state,
      discriminator_params=discriminator_params,
      discriminator_state=new_discriminator_state,
      policy_params=state.policy_params,  # Not modified.
      key=key,
      steps=state.steps + 1,
  )
  return new_state, loss


class AILSample(NamedTuple):
  discriminator_sample: types.Transition
  direct_sample: reverb.ReplaySample
  demonstration_sample: types.Transition


class AILLearner(acme.Learner):
  """AIL learner."""

  def __init__(
      self,
      counter: counting.Counter,
      direct_rl_learner_factory: Callable[[Iterator[reverb.ReplaySample]],
                                          acme.Learner],
      loss_fn: losses.Loss,
      iterator: Iterator[AILSample],
      discriminator_optimizer: optax.GradientTransformation,
      ail_network: ail_networks.AILNetworks,
      discriminator_key: networks_lib.PRNGKey,
      is_sequence_based: bool,
      num_sgd_steps_per_step: int = 1,
      policy_variable_name: Optional[str] = None,
      logger: Optional[loggers.Logger] = None):
    """AIL Learner.

    Args:
      counter: Counter.
      direct_rl_learner_factory: Function that creates the direct RL learner
        when passed a replay sample iterator.
      loss_fn: Discriminator loss.
      iterator: Iterator that returns AILSamples.
      discriminator_optimizer: Discriminator optax optimizer.
      ail_network: AIL networks.
      discriminator_key: RNG key.
      is_sequence_based: If True, a direct rl algorithm is using SequenceAdder
        data format. Otherwise the learner assumes that the direct rl algorithm
        is using NStepTransitionAdder.
      num_sgd_steps_per_step: Number of discriminator gradient updates per step.
      policy_variable_name: The name of the policy variable to retrieve
        direct_rl policy parameters.
      logger: Logger.
    """
    self._is_sequence_based = is_sequence_based

    state_key, networks_key = jax.random.split(discriminator_key)

    # Generator expression that works the same as an iterator.
    # https://pymbook.readthedocs.io/en/latest/igd.html#generator-expressions
    iterator, direct_rl_iterator = itertools.tee(iterator)
    direct_rl_iterator = (
        self._process_sample(sample.direct_sample)
        for sample in direct_rl_iterator)
    self._direct_rl_learner = direct_rl_learner_factory(direct_rl_iterator)

    self._iterator = iterator

    if policy_variable_name is not None:

      def get_policy_params():
        return self._direct_rl_learner.get_variables([policy_variable_name])[0]

      self._get_policy_params = get_policy_params

    else:
      self._get_policy_params = lambda: None

    # General learner book-keeping and loggers.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger(
        'learner', asynchronous=True, serialize_fn=utils.fetch_devicearray)

    # Use the JIT compiler.
    self._update_step = functools.partial(
        ail_update_step,
        optimizer=discriminator_optimizer,
        ail_network=ail_network,
        loss_fn=loss_fn)
    self._update_step = utils.process_multiple_batches(self._update_step,
                                                       num_sgd_steps_per_step)
    self._update_step = jax.jit(self._update_step)

    discriminator_params, discriminator_state = (
        ail_network.discriminator_network.init(networks_key))
    self._state = DiscriminatorTrainingState(
        optimizer_state=discriminator_optimizer.init(discriminator_params),
        discriminator_params=discriminator_params,
        discriminator_state=discriminator_state,
        policy_params=self._get_policy_params(),
        key=state_key,
        steps=0,
    )

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

    self._get_reward = jax.jit(
        functools.partial(
            ail_networks.compute_ail_reward, networks=ail_network))

  def _process_sample(self, sample: reverb.ReplaySample) -> reverb.ReplaySample:
    """Updates the reward of the replay sample.

    Args:
      sample: Replay sample to update the reward to.

    Returns:
      The replay sample with an updated reward.
    """
    transitions = reverb_utils.replay_sample_to_sars_transition(
        sample, is_sequence=self._is_sequence_based)
    rewards = self._get_reward(self._state.discriminator_params,
                               self._state.discriminator_state,
                               self._state.policy_params, transitions)

    return sample._replace(data=sample.data._replace(reward=rewards))

  def step(self):
    sample = next(self._iterator)
    rb_transitions = sample.discriminator_sample
    demo_transitions = sample.demonstration_sample

    if demo_transitions.reward.shape != rb_transitions.reward.shape:
      raise ValueError(
          'Different shapes for demo transitions and rb_transitions: '
          f'{demo_transitions.reward.shape} != {rb_transitions.reward.shape}')

    # Update the parameters of the policy before doing a gradient step.
    state = self._state._replace(policy_params=self._get_policy_params())
    self._state, metrics = self._update_step(state,
                                             (demo_transitions, rb_transitions))

    # The order is important for AIRL.
    # In AIRL, the discriminator update depends on the logpi of the direct rl
    # policy.
    # When updating the discriminator, we want the logpi for which the
    # transitions were made with and not an updated one.
    # Get data from replay (dropping extras if any). Note there is no
    # extra data here because we do not insert any into Reverb.
    self._direct_rl_learner.step()

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Increment counts and record the current time.
    counts = self._counter.increment(steps=1, walltime=elapsed_time)

    # Attempts to write the logs.
    self._logger.write({**metrics, **counts})

  def get_variables(self, names: List[str]) -> List[Any]:
    rewarder_dict = {'discriminator': self._state.discriminator_params}

    learner_names = [name for name in names if name not in rewarder_dict]
    learner_dict = {}
    if learner_names:
      learner_dict = dict(
          zip(learner_names,
              self._direct_rl_learner.get_variables(learner_names)))

    variables = [
        rewarder_dict.get(name, learner_dict.get(name, None)) for name in names
    ]
    return variables

  def save(self) -> TrainingState:
    return TrainingState(
        rewarder_state=self._state,
        learner_state=self._direct_rl_learner.save())

  def restore(self, state: TrainingState):
    self._state = state.rewarder_state
    self._direct_rl_learner.restore(state.learner_state)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the D4PG agent class."""

import copy
from typing import Callable, Dict, Optional

import acme
from acme import specs
from acme.agents.tf.d4pg import agent
from acme.tf import savers as tf2_savers
from acme.utils import counting
from acme.utils import loggers
from acme.utils import lp_utils
import dm_env
import launchpad as lp
import reverb
import sonnet as snt
import tensorflow as tf


class DistributedD4PG:
  """Program definition for D4PG."""

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      network_factory: Callable[[specs.BoundedArray], Dict[str, snt.Module]],
      num_actors: int = 1,
      num_caches: int = 0,
      environment_spec: Optional[specs.EnvironmentSpec] = None,
      batch_size: int = 256,
      prefetch_size: int = 4,
      min_replay_size: int = 1000,
      max_replay_size: int = 1000000,
      samples_per_insert: Optional[float] = 32.0,
      n_step: int = 5,
      sigma: float = 0.3,
      clipping: bool = True,
      discount: float = 0.99,
      policy_optimizer: Optional[snt.Optimizer] = None,
      critic_optimizer: Optional[snt.Optimizer] = None,
      target_update_period: int = 100,
      max_actor_steps: Optional[int] = None,
      log_every: float = 10.0,
  ):

    if not environment_spec:
      environment_spec = specs.make_environment_spec(environment_factory(False))

    # TODO(mwhoffman): Make network_factory directly return the struct.
    # TODO(mwhoffman): Make the factory take the entire spec.
    def wrapped_network_factory(action_spec):
      networks_dict = network_factory(action_spec)
      networks = agent.D4PGNetworks(
          policy_network=networks_dict.get('policy'),
          critic_network=networks_dict.get('critic'),
          observation_network=networks_dict.get('observation', tf.identity))
      return networks

    self._environment_factory = environment_factory
    self._network_factory = wrapped_network_factory
    self._environment_spec = environment_spec
    self._sigma = sigma
    self._num_actors = num_actors
    self._num_caches = num_caches
    self._max_actor_steps = max_actor_steps
    self._log_every = log_every

    self._builder = agent.D4PGBuilder(
        # TODO(mwhoffman): pass the config dataclass in directly.
        # TODO(mwhoffman): use the limiter rather than the workaround below.
        agent.D4PGConfig(
            discount=discount,
            batch_size=batch_size,
            prefetch_size=prefetch_size,
            target_update_period=target_update_period,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            min_replay_size=min_replay_size,
            max_replay_size=max_replay_size,
            samples_per_insert=samples_per_insert,
            n_step=n_step,
            sigma=sigma,
            clipping=clipping,
        ))

  def replay(self):
    """The replay storage."""
    return self._builder.make_replay_tables(self._environment_spec)

  def counter(self):
    return tf2_savers.CheckpointingRunner(counting.Counter(),
                                          time_delta_minutes=1,
                                          subdirectory='counter')

  def coordinator(self, counter: counting.Counter):
    return lp_utils.StepsLimiter(counter, self._max_actor_steps)

  def learner(
      self,
      replay: reverb.Client,
      counter: counting.Counter,
  ):
    """The Learning part of the agent."""

    # Create the networks to optimize (online) and target networks.
    online_networks = self._network_factory(self._environment_spec.actions)
    target_networks = copy.deepcopy(online_networks)

    # Initialize the networks.
    online_networks.init(self._environment_spec)
    target_networks.init(self._environment_spec)

    dataset = self._builder.make_dataset_iterator(replay)
    counter = counting.Counter(counter, 'learner')
    logger = loggers.make_default_logger(
        'learner', time_delta=self._log_every, steps_key='learner_steps')

    return self._builder.make_learner(
        networks=(online_networks, target_networks),
        dataset=dataset,
        counter=counter,
        logger=logger,
        checkpoint=True,
    )

  def actor(
      self,
      replay: reverb.Client,
      variable_source: acme.VariableSource,
      counter: counting.Counter,
  ) -> acme.EnvironmentLoop:
    """The actor process."""

    # Create the behavior policy.
    networks = self._network_factory(self._environment_spec.actions)
    networks.init(self._environment_spec)
    policy_network = networks.make_policy(
        environment_spec=self._environment_spec,
        sigma=self._sigma,
    )

    # Create the agent.
    actor = self._builder.make_actor(
        policy_network=policy_network,
        adder=self._builder.make_adder(replay),
        variable_source=variable_source,
    )

    # Create the environment.
    environment = self._environment_factory(False)

    # Create logger and counter; actors will not spam bigtable.
    counter = counting.Counter(counter, 'actor')
    logger = loggers.make_default_logger(
        'actor',
        save_data=False,
        time_delta=self._log_every,
        steps_key='actor_steps')

    # Create the loop to connect environment and agent.
    return acme.EnvironmentLoop(environment, actor, counter, logger)

  def evaluator(
      self,
      variable_source: acme.VariableSource,
      counter: counting.Counter,
      logger: Optional[loggers.Logger] = None,
  ):
    """The evaluation process."""

    # Create the behavior policy.
    networks = self._network_factory(self._environment_spec.actions)
    networks.init(self._environment_spec)
    policy_network = networks.make_policy(self._environment_spec)

    # Create the agent.
    actor = self._builder.make_actor(
        policy_network=policy_network,
        variable_source=variable_source,
    )

    # Make the environment.
    environment = self._environment_factory(True)

    # Create logger and counter.
    counter = counting.Counter(counter, 'evaluator')
    logger = logger or loggers.make_default_logger(
        'evaluator',
        time_delta=self._log_every,
        steps_key='evaluator_steps',
    )

    # Create the run loop and return it.
    return acme.EnvironmentLoop(environment, actor, counter, logger)

  def build(self, name='d4pg'):
    """Build the distributed agent topology."""
    program = lp.Program(name=name)

    with program.group('replay'):
      replay = program.add_node(lp.ReverbNode(self.replay))

    with program.group('counter'):
      counter = program.add_node(lp.CourierNode(self.counter))

    if self._max_actor_steps:
      with program.group('coordinator'):
        _ = program.add_node(lp.CourierNode(self.coordinator, counter))

    with program.group('learner'):
      learner = program.add_node(lp.CourierNode(self.learner, replay, counter))

    with program.group('evaluator'):
      program.add_node(lp.CourierNode(self.evaluator, learner, counter))

    if not self._num_caches:
      # Use our learner as a single variable source.
      sources = [learner]
    else:
      with program.group('cacher'):
        # Create a set of learner caches.
        sources = []
        for _ in range(self._num_caches):
          cacher = program.add_node(
              lp.CacherNode(
                  learner, refresh_interval_ms=2000, stale_after_ms=4000))
          sources.append(cacher)

    with program.group('actor'):
      # Add actors which pull round-robin from our variable sources.
      for actor_id in range(self._num_actors):
        source = sources[actor_id % len(sources)]
        program.add_node(lp.CourierNode(self.actor, replay, source, counter))

    return program



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""IMPALA networks definition."""

import dataclasses
from typing import Any, Generic, Optional, Tuple

from acme import specs
from acme.agents.jax.impala import types
from acme.jax import networks as networks_lib
from acme.jax import utils
import haiku as hk


@dataclasses.dataclass
class IMPALANetworks(Generic[types.RecurrentState]):

  """Pure functions representing IMPALA's recurrent network components.

  Attributes:
    forward_fn: Selects next action using the network at the given recurrent
      state.

    unroll_init_fn: Initializes params for forward_fn and unroll_fn.

    unroll_fn: Applies the unrolled network to a sequence of observations, for
      learning.
    initial_state_init_fn: Initializes params for initial_state_fn.
    initial_state_fn: Recurrent state at the beginning of an episode.
  """
  forward_fn: types.PolicyValueFn
  unroll_init_fn: types.PolicyValueInitFn
  unroll_fn: types.PolicyValueFn
  initial_state_init_fn: types.RecurrentStateInitFn
  initial_state_fn: types.RecurrentStateFn


def make_haiku_networks(
    env_spec: specs.EnvironmentSpec,
    forward_fn: Any,
    initial_state_fn: Any,
    unroll_fn: Any) -> IMPALANetworks[types.RecurrentState]:
  """Builds functional impala network from recurrent model definitions."""
  # Make networks purely functional.
  forward_hk = hk.without_apply_rng(hk.transform(forward_fn))
  initial_state_hk = hk.without_apply_rng(hk.transform(initial_state_fn))
  unroll_hk = hk.without_apply_rng(hk.transform(unroll_fn))

  # Define networks init functions.
  def initial_state_init_fn(rng: networks_lib.PRNGKey) -> hk.Params:
    return initial_state_hk.init(rng)
  # Note: batch axis is not needed for the actors.
  dummy_obs = utils.zeros_like(env_spec.observations)
  dummy_obs_sequence = utils.add_batch_dim(dummy_obs)
  def unroll_init_fn(
      rng: networks_lib.PRNGKey,
      initial_state: types.RecurrentState) -> hk.Params:
    return unroll_hk.init(rng, dummy_obs_sequence, initial_state)

  return IMPALANetworks(
      forward_fn=forward_hk.apply,
      unroll_init_fn=unroll_init_fn,
      unroll_fn=unroll_hk.apply,
      initial_state_init_fn=initial_state_init_fn,
      initial_state_fn=initial_state_hk.apply)


HaikuLSTMOutputs = Tuple[Tuple[networks_lib.Logits, networks_lib.Value],
                         hk.LSTMState]


def make_atari_networks(env_spec: specs.EnvironmentSpec
                        ) -> IMPALANetworks[hk.LSTMState]:
  """Builds default IMPALA networks for Atari games."""

  def forward_fn(inputs: types.Observation, state: hk.LSTMState
                 ) -> HaikuLSTMOutputs:
    model = networks_lib.DeepIMPALAAtariNetwork(env_spec.actions.num_values)
    return model(inputs, state)

  def initial_state_fn(batch_size: Optional[int] = None) -> hk.LSTMState:
    model = networks_lib.DeepIMPALAAtariNetwork(env_spec.actions.num_values)
    return model.initial_state(batch_size)

  def unroll_fn(inputs: types.Observation, state: hk.LSTMState
                ) -> HaikuLSTMOutputs:
    model = networks_lib.DeepIMPALAAtariNetwork(env_spec.actions.num_values)
    return model.unroll(inputs, state)

  return make_haiku_networks(
      env_spec=env_spec, forward_fn=forward_fn,
      initial_state_fn=initial_state_fn, unroll_fn=unroll_fn)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for OpenSpiel environment loop."""

import unittest
from absl.testing import absltest
from absl.testing import parameterized

import acme
from acme import core
from acme import specs
from acme import types
from acme import wrappers

import dm_env
import numpy as np
import tree

SKIP_OPEN_SPIEL_TESTS = False
SKIP_OPEN_SPIEL_MESSAGE = 'open_spiel not installed.'

try:
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  from acme.environment_loops import open_spiel_environment_loop
  from acme.wrappers import open_spiel_wrapper
  from open_spiel.python import rl_environment
  # pytype: disable=import-error

  class RandomActor(core.Actor):
    """Fake actor which generates random actions and validates specs."""

    def __init__(self, spec: specs.EnvironmentSpec):
      self._spec = spec
      self.num_updates = 0

    def select_action(self, observation: open_spiel_wrapper.OLT) -> int:
      _validate_spec(self._spec.observations, observation)
      legals = np.array(np.nonzero(observation.legal_actions), dtype=np.int32)
      return np.random.choice(legals[0])

    def observe_first(self, timestep: dm_env.TimeStep):
      _validate_spec(self._spec.observations, timestep.observation)

    def observe(self, action: types.NestedArray,
                next_timestep: dm_env.TimeStep):
      _validate_spec(self._spec.actions, action)
      _validate_spec(self._spec.rewards, next_timestep.reward)
      _validate_spec(self._spec.discounts, next_timestep.discount)
      _validate_spec(self._spec.observations, next_timestep.observation)

    def update(self, wait: bool = False):
      self.num_updates += 1

except ModuleNotFoundError:
  SKIP_OPEN_SPIEL_TESTS = True


def _validate_spec(spec: types.NestedSpec, value: types.NestedArray):
  """Validate a value from a potentially nested spec."""
  tree.assert_same_structure(value, spec)
  tree.map_structure(lambda s, v: s.validate(v), spec, value)


@unittest.skipIf(SKIP_OPEN_SPIEL_TESTS, SKIP_OPEN_SPIEL_MESSAGE)
class OpenSpielEnvironmentLoopTest(parameterized.TestCase):

  def test_loop_run(self):
    raw_env = rl_environment.Environment('tic_tac_toe')
    env = open_spiel_wrapper.OpenSpielWrapper(raw_env)
    env = wrappers.SinglePrecisionWrapper(env)
    environment_spec = acme.make_environment_spec(env)

    actors = []
    for _ in range(env.num_players):
      actors.append(RandomActor(environment_spec))

    loop = open_spiel_environment_loop.OpenSpielEnvironmentLoop(env, actors)
    result = loop.run_episode()
    self.assertIn('episode_length', result)
    self.assertIn('episode_return', result)
    self.assertIn('steps_per_second', result)

    loop.run(num_episodes=10)
    loop.run(num_steps=100)


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An image logger, for writing out arrays to disk as PNG."""

import collections
import pathlib
from typing import Optional

from absl import logging
from acme.utils.loggers import base
from PIL import Image


class ImageLogger(base.Logger):
  """Logger for writing NumPy arrays as PNG images to disk.

  Assumes that all data passed are NumPy arrays that can be converted to images.

  TODO(jaslanides): Make this stateless/robust to preemptions.
  """

  def __init__(
      self,
      directory: str,
      *,
      label: str = '',
      mode: Optional[str] = None,
  ):
    """Initialises the writer.

    Args:
      directory: Base directory to which images are logged.
      label: Optional subdirectory in which to save images.
      mode: Image mode for use with Pillow. If `None` (default), mode is
        determined by data type. See [0] for details.

    [0] https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes
    """

    self._path = self._get_path(directory, label)
    if not self._path.exists():
      self._path.mkdir(parents=True)

    self._mode = mode
    self._indices = collections.defaultdict(int)

  def write(self, data: base.LoggingData):
    for k, v in data.items():
      image = Image.fromarray(v, mode=self._mode)
      path = self._path / f'{k}_{self._indices[k]:06}.png'
      self._indices[k] += 1
      with path.open(mode='wb') as f:
        logging.info('Writing image to %s.', str(path))
        image.save(f)

  def close(self):
    pass

  @property
  def directory(self) -> str:
    return str(self._path)

  def _get_path(self, *args, **kwargs) -> pathlib.Path:
    return pathlib.Path(*args, **kwargs)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""D4PG agent implementation."""

import copy
import dataclasses
from typing import Iterator, List, Optional, Tuple

from acme import adders
from acme import core
from acme import datasets
from acme import specs
from acme import types
from acme.adders import reverb as reverb_adders
from acme.agents import agent
from acme.agents.tf import actors
from acme.agents.tf.d4pg import learning
from acme.tf import networks as network_utils
from acme.tf import utils
from acme.tf import variable_utils
from acme.utils import counting
from acme.utils import loggers
import reverb
import sonnet as snt
import tensorflow as tf


@dataclasses.dataclass
class D4PGConfig:
  """Configuration options for the D4PG agent."""

  discount: float = 0.99
  batch_size: int = 256
  prefetch_size: int = 4
  target_update_period: int = 100
  policy_optimizer: Optional[snt.Optimizer] = None
  critic_optimizer: Optional[snt.Optimizer] = None
  min_replay_size: int = 1000
  max_replay_size: int = 1000000
  samples_per_insert: Optional[float] = 32.0
  n_step: int = 5
  sigma: float = 0.3
  clipping: bool = True
  replay_table_name: str = reverb_adders.DEFAULT_PRIORITY_TABLE


@dataclasses.dataclass
class D4PGNetworks:
  """Structure containing the networks for D4PG."""

  policy_network: snt.Module
  critic_network: snt.Module
  observation_network: snt.Module

  def __init__(
      self,
      policy_network: snt.Module,
      critic_network: snt.Module,
      observation_network: types.TensorTransformation,
  ):
    # This method is implemented (rather than added by the dataclass decorator)
    # in order to allow observation network to be passed as an arbitrary tensor
    # transformation rather than as a snt Module.
    # TODO(mwhoffman): use Protocol rather than Module/TensorTransformation.
    self.policy_network = policy_network
    self.critic_network = critic_network
    self.observation_network = utils.to_sonnet_module(observation_network)

  def init(self, environment_spec: specs.EnvironmentSpec):
    """Initialize the networks given an environment spec."""
    # Get observation and action specs.
    act_spec = environment_spec.actions
    obs_spec = environment_spec.observations

    # Create variables for the observation net and, as a side-effect, get a
    # spec describing the embedding space.
    emb_spec = utils.create_variables(self.observation_network, [obs_spec])

    # Create variables for the policy and critic nets.
    _ = utils.create_variables(self.policy_network, [emb_spec])
    _ = utils.create_variables(self.critic_network, [emb_spec, act_spec])

  def make_policy(
      self,
      environment_spec: specs.EnvironmentSpec,
      sigma: float = 0.0,
  ) -> snt.Module:
    """Create a single network which evaluates the policy."""
    # Stack the observation and policy networks.
    stack = [
        self.observation_network,
        self.policy_network,
    ]

    # If a stochastic/non-greedy policy is requested, add Gaussian noise on
    # top to enable a simple form of exploration.
    # TODO(mwhoffman): Refactor this to remove it from the class.
    if sigma > 0.0:
      stack += [
          network_utils.ClippedGaussian(sigma),
          network_utils.ClipToSpec(environment_spec.actions),
      ]

    # Return a network which sequentially evaluates everything in the stack.
    return snt.Sequential(stack)


class D4PGBuilder:
  """Builder for D4PG which constructs individual components of the agent."""

  def __init__(self, config: D4PGConfig):
    self._config = config

  def make_replay_tables(
      self,
      environment_spec: specs.EnvironmentSpec,
  ) -> List[reverb.Table]:
    """Create tables to insert data into."""
    if self._config.samples_per_insert is None:
      # We will take a samples_per_insert ratio of None to mean that there is
      # no limit, i.e. this only implies a min size limit.
      limiter = reverb.rate_limiters.MinSize(self._config.min_replay_size)

    else:
      # Create enough of an error buffer to give a 10% tolerance in rate.
      samples_per_insert_tolerance = 0.1 * self._config.samples_per_insert
      error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
      limiter = reverb.rate_limiters.SampleToInsertRatio(
          min_size_to_sample=self._config.min_replay_size,
          samples_per_insert=self._config.samples_per_insert,
          error_buffer=error_buffer)

    replay_table = reverb.Table(
        name=self._config.replay_table_name,
        sampler=reverb.selectors.Uniform(),
        remover=reverb.selectors.Fifo(),
        max_size=self._config.max_replay_size,
        rate_limiter=limiter,
        signature=reverb_adders.NStepTransitionAdder.signature(
            environment_spec))

    return [replay_table]

  def make_dataset_iterator(
      self,
      reverb_client: reverb.Client,
  ) -> Iterator[reverb.ReplaySample]:
    """Create a dataset iterator to use for learning/updating the agent."""
    # The dataset provides an interface to sample from replay.
    dataset = datasets.make_reverb_dataset(
        table=self._config.replay_table_name,
        server_address=reverb_client.server_address,
        batch_size=self._config.batch_size,
        prefetch_size=self._config.prefetch_size)

    # TODO(b/155086959): Fix type stubs and remove.
    return iter(dataset)  # pytype: disable=wrong-arg-types

  def make_adder(
      self,
      replay_client: reverb.Client,
  ) -> adders.Adder:
    """Create an adder which records data generated by the actor/environment."""
    return reverb_adders.NStepTransitionAdder(
        priority_fns={self._config.replay_table_name: lambda x: 1.},
        client=replay_client,
        n_step=self._config.n_step,
        discount=self._config.discount)

  def make_actor(
      self,
      policy_network: snt.Module,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None,
  ):
    """Create an actor instance."""
    if variable_source:
      # Create the variable client responsible for keeping the actor up-to-date.
      variable_client = variable_utils.VariableClient(
          client=variable_source,
          variables={'policy': policy_network.variables},
          update_period=1000,
      )

      # Make sure not to use a random policy after checkpoint restoration by
      # assigning variables before running the environment loop.
      variable_client.update_and_wait()

    else:
      variable_client = None

    # Create the actor which defines how we take actions.
    return actors.FeedForwardActor(
        policy_network=policy_network,
        adder=adder,
        variable_client=variable_client,
    )

  def make_learner(
      self,
      networks: Tuple[D4PGNetworks, D4PGNetworks],
      dataset: Iterator[reverb.ReplaySample],
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      checkpoint: bool = False,
  ):
    """Creates an instance of the learner."""
    online_networks, target_networks = networks

    # The learner updates the parameters (and initializes them).
    return learning.D4PGLearner(
        policy_network=online_networks.policy_network,
        critic_network=online_networks.critic_network,
        observation_network=online_networks.observation_network,
        target_policy_network=target_networks.policy_network,
        target_critic_network=target_networks.critic_network,
        target_observation_network=target_networks.observation_network,
        policy_optimizer=self._config.policy_optimizer,
        critic_optimizer=self._config.critic_optimizer,
        clipping=self._config.clipping,
        discount=self._config.discount,
        target_update_period=self._config.target_update_period,
        dataset_iterator=dataset,
        counter=counter,
        logger=logger,
        checkpoint=checkpoint,
    )


class D4PG(agent.Agent):
  """D4PG Agent.

  This implements a single-process D4PG agent. This is an actor-critic algorithm
  that generates data via a behavior policy, inserts N-step transitions into
  a replay buffer, and periodically updates the policy (and as a result the
  behavior) by sampling uniformly from this buffer.
  """

  def __init__(
      self,
      environment_spec: specs.EnvironmentSpec,
      policy_network: snt.Module,
      critic_network: snt.Module,
      observation_network: types.TensorTransformation = tf.identity,
      discount: float = 0.99,
      batch_size: int = 256,
      prefetch_size: int = 4,
      target_update_period: int = 100,
      policy_optimizer: Optional[snt.Optimizer] = None,
      critic_optimizer: Optional[snt.Optimizer] = None,
      min_replay_size: int = 1000,
      max_replay_size: int = 1000000,
      samples_per_insert: float = 32.0,
      n_step: int = 5,
      sigma: float = 0.3,
      clipping: bool = True,
      replay_table_name: str = reverb_adders.DEFAULT_PRIORITY_TABLE,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      checkpoint: bool = True,
  ):
    """Initialize the agent.

    Args:
      environment_spec: description of the actions, observations, etc.
      policy_network: the online (optimized) policy.
      critic_network: the online critic.
      observation_network: optional network to transform the observations before
        they are fed into any network.
      discount: discount to use for TD updates.
      batch_size: batch size for updates.
      prefetch_size: size to prefetch from replay.
      target_update_period: number of learner steps to perform before updating
        the target networks.
      policy_optimizer: optimizer for the policy network updates.
      critic_optimizer: optimizer for the critic network updates.
      min_replay_size: minimum replay size before updating.
      max_replay_size: maximum replay size.
      samples_per_insert: number of samples to take from replay for every insert
        that is made.
      n_step: number of steps to squash into a single transition.
      sigma: standard deviation of zero-mean, Gaussian exploration noise.
      clipping: whether to clip gradients by global norm.
      replay_table_name: string indicating what name to give the replay table.
      counter: counter object used to keep track of steps.
      logger: logger object to be used by learner.
      checkpoint: boolean indicating whether to checkpoint the learner.
    """
    # Create the Builder object which will internally create agent components.
    builder = D4PGBuilder(
        # TODO(mwhoffman): pass the config dataclass in directly.
        # TODO(mwhoffman): use the limiter rather than the workaround below.
        # Right now this modifies min_replay_size and samples_per_insert so that
        # they are not controlled by a limiter and are instead handled by the
        # Agent base class (the above TODO directly references this behavior).
        D4PGConfig(
            discount=discount,
            batch_size=batch_size,
            prefetch_size=prefetch_size,
            target_update_period=target_update_period,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            min_replay_size=1,  # Let the Agent class handle this.
            max_replay_size=max_replay_size,
            samples_per_insert=None,  # Let the Agent class handle this.
            n_step=n_step,
            sigma=sigma,
            clipping=clipping,
            replay_table_name=replay_table_name,
        ))

    # TODO(mwhoffman): pass the network dataclass in directly.
    online_networks = D4PGNetworks(policy_network=policy_network,
                                   critic_network=critic_network,
                                   observation_network=observation_network)

    # Target networks are just a copy of the online networks.
    target_networks = copy.deepcopy(online_networks)

    # Initialize the networks.
    online_networks.init(environment_spec)
    target_networks.init(environment_spec)

    # TODO(mwhoffman): either make this Dataclass or pass only one struct.
    # The network struct passed to make_learner is just a tuple for the
    # time-being (for backwards compatibility).
    networks = (online_networks, target_networks)

    # Create the behavior policy.
    policy_network = online_networks.make_policy(environment_spec, sigma)

    # Create the replay server and grab its address.
    replay_tables = builder.make_replay_tables(environment_spec)
    replay_server = reverb.Server(replay_tables, port=None)
    replay_client = reverb.Client(f'localhost:{replay_server.port}')

    # Create actor, dataset, and learner for generating, storing, and consuming
    # data respectively.
    adder = builder.make_adder(replay_client)
    actor = builder.make_actor(policy_network, adder)
    dataset = builder.make_dataset_iterator(replay_client)
    learner = builder.make_learner(networks, dataset, counter, logger,
                                   checkpoint)

    super().__init__(
        actor=actor,
        learner=learner,
        min_observations=max(batch_size, min_replay_size),
        observations_per_step=float(batch_size) / samples_per_insert)

    # Save the replay so we don't garbage collect it.
    self._replay_server = replay_server



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Package metadata for acme.

This is kept in a separate module so that it can be imported from setup.py, at
a time when acme's dependencies may not have been installed yet.
"""

# We follow Semantic Versioning (https://semver.org/)
_MAJOR_VERSION = '0'
_MINOR_VERSION = '3'
_PATCH_VERSION = '0'

# Example: '0.4.2'
__version__ = '.'.join([_MAJOR_VERSION, _MINOR_VERSION, _PATCH_VERSION])



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the AIL agent."""
import functools
import shutil
from typing import Any, Callable

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized
import acme
from acme import specs
from acme import types
from acme.agents.jax import ail
from acme.agents.jax import ppo
from acme.agents.jax import sac
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.jax.layouts import local_layout
from acme.testing import fakes
from acme.utils import counting
from flax import linen
import haiku as hk
import jax
from jax import numpy as jnp
import numpy as np


NUM_DISCRETE_ACTIONS = 5
NUM_OBSERVATIONS = 10
OBS_SHAPE = (10, 5)
OBS_DTYPE = np.float32
EPISODE_LENGTH = 10
CONTINUOUS_ACTION_DIM = 3
CONTINUOUS_OBS_DIM = 5


def make_sac_logpi(
    sac_networks: sac.SACNetworks
) -> Callable[
    [networks_lib.Params, networks_lib.Observation, networks_lib.Action],
    jnp.ndarray]:
  """Returns the logpi function for SAC.

  Args:
    sac_networks: sac networks.

  Returns:
    A function from model params, obsevations, actions to log probability.
  """

  def logpi(params: networks_lib.Params, observations: networks_lib.Observation,
            actions: networks_lib.Action) -> jnp.ndarray:
    """Log probability of the action with the current policy."""
    model_output = sac_networks.policy_network.apply(params, observations)
    return sac_networks.log_prob(model_output, actions)

  return logpi


def ppo_forward_fn(inputs, num_actions):
  policy_network = hk.Sequential([
      utils.batch_concat,
      hk.nets.MLP([64, 64]),
      networks_lib.CategoricalHead(num_actions)
  ])
  value_network = hk.Sequential([
      utils.batch_concat,
      hk.nets.MLP([64, 64]),
      hk.Linear(1), lambda x: jnp.squeeze(x, axis=-1)
  ])

  action_distribution = policy_network(inputs)
  value = value_network(inputs)
  return (action_distribution, value)


def make_ppo_networks(
    spec: specs.EnvironmentSpec) -> networks_lib.FeedForwardNetwork:
  """Creates Haiku networks to be used by the agent."""

  num_actions = spec.actions.num_values

  forward_fn = functools.partial(ppo_forward_fn, num_actions=num_actions)

  # Transform into pure functions.
  forward_fn = hk.without_apply_rng(hk.transform(forward_fn))

  dummy_obs = utils.zeros_like(spec.observations)
  dummy_obs = utils.add_batch_dim(dummy_obs)  # Dummy 'sequence' dim.
  return networks_lib.FeedForwardNetwork(
      lambda rng: forward_fn.init(rng, dummy_obs), forward_fn.apply)


class DiscriminatorModule(linen.Module):
  """Discriminator module that concatenates its inputs."""
  # Environment spec
  environment_spec: specs.EnvironmentSpec
  # Network core
  network_core: Callable[..., Any]

  @linen.compact
  def __call__(self, observations: jnp.ndarray, actions: jnp.ndarray,
               next_observations: jnp.ndarray, is_training: bool,
               rng: networks_lib.PRNGKey):
    output = self.network_core(observations)
    output = jnp.squeeze(output, axis=-1)
    return output


class AILTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('sac', 'sac'), ('ppo', 'ppo'), ('sac_airl', 'sac', True),
      ('sac_logpi', 'sac', False, True),
      ('sac_dropout', 'sac', False, False, .75),
      ('sac_spectral_norm', 'sac', False, True, 0., 1.))
  def test_ail(self,
               algo,
               airl_discriminator=False,
               subtract_logpi=False,
               dropout=0.,
               lipschitz_coeff=None):
    shutil.rmtree(flags.FLAGS.test_tmpdir, ignore_errors=True)
    batch_size = 8
    # Mujoco environment and associated demonstration dataset.
    if algo == 'ppo':
      environment = fakes.DiscreteEnvironment(
          num_actions=NUM_DISCRETE_ACTIONS,
          num_observations=NUM_OBSERVATIONS,
          obs_shape=OBS_SHAPE,
          obs_dtype=OBS_DTYPE,
          episode_length=EPISODE_LENGTH)
    else:
      environment = fakes.ContinuousEnvironment(
          episode_length=EPISODE_LENGTH,
          action_dim=CONTINUOUS_ACTION_DIM,
          observation_dim=CONTINUOUS_OBS_DIM,
          bounded=True)
    spec = specs.make_environment_spec(environment)

    if algo == 'sac':
      networks = sac.make_networks(spec=spec)
      config = sac.SACConfig(batch_size=batch_size,
                             samples_per_insert_tolerance_rate=float('inf'),
                             min_replay_size=1)
      base_builder = sac.SACBuilder(config=config)
      direct_rl_batch_size = batch_size
      behavior_policy = sac.apply_policy_and_sample(networks)
    elif algo == 'ppo':
      unroll_length = 5
      distribution_value_networks = make_ppo_networks(spec)
      networks = ppo.make_ppo_networks(distribution_value_networks)
      config = ppo.PPOConfig(
          unroll_length=unroll_length,
          num_minibatches=2,
          num_epochs=4,
          batch_size=batch_size)
      base_builder = ppo.PPOBuilder(config=config)
      direct_rl_batch_size = batch_size * unroll_length
      behavior_policy = jax.jit(ppo.make_inference_fn(networks),
                                backend='cpu')
    else:
      raise ValueError(f'Unexpected algorithm {algo}')

    if subtract_logpi:
      assert algo == 'sac'
      logpi_fn = make_sac_logpi(networks)
    else:
      logpi_fn = None

    if algo == 'ppo':
      embedding = lambda x: jnp.reshape(x, list(x.shape[:-2]) + [-1])
    else:
      embedding = lambda x: x

    def discriminator(*args, **kwargs) -> networks_lib.Logits:
      if airl_discriminator:
        return ail.AIRLModule(
            environment_spec=spec,
            use_action=True,
            use_next_obs=True,
            discount=.99,
            g_core=ail.DiscriminatorMLP(
                [4, 4],
                hidden_dropout_rate=dropout,
                spectral_normalization_lipschitz_coeff=lipschitz_coeff
            ),
            h_core=ail.DiscriminatorMLP(
                [4, 4],
                hidden_dropout_rate=dropout,
                spectral_normalization_lipschitz_coeff=lipschitz_coeff
            ),
            observation_embedding=embedding)(*args, **kwargs)
      else:
        return ail.DiscriminatorModule(
            environment_spec=spec,
            use_action=True,
            use_next_obs=True,
            network_core=ail.DiscriminatorMLP(
                [4, 4],
                hidden_dropout_rate=dropout,
                spectral_normalization_lipschitz_coeff=lipschitz_coeff
            ),
            observation_embedding=embedding)(*args, **kwargs)

    discriminator_transformed = hk.without_apply_rng(
        hk.transform_with_state(discriminator))

    discriminator_network = ail.make_discriminator(
        environment_spec=spec,
        discriminator_transformed=discriminator_transformed,
        logpi_fn=logpi_fn)

    networks = ail.AILNetworks(discriminator_network, lambda x: x, networks)

    builder = ail.AILBuilder(
        base_builder,
        config=ail.AILConfig(
            is_sequence_based=(algo == 'ppo'),
            share_iterator=True,
            direct_rl_batch_size=direct_rl_batch_size,
            discriminator_batch_size=2,
            policy_variable_name='policy' if subtract_logpi else None,
            min_replay_size=1),
        discriminator_loss=ail.losses.gail_loss(),
        make_demonstrations=fakes.transition_iterator(environment))

    # Construct the agent.
    agent = local_layout.LocalLayout(
        seed=0,
        environment_spec=spec,
        builder=builder,
        networks=networks,
        policy_network=behavior_policy,
        min_replay_size=1,
        batch_size=batch_size)

    # Train the agent.
    train_loop = acme.EnvironmentLoop(environment, agent)
    train_loop.run(num_episodes=(10 if algo == 'ppo' else 1))

  def test_ail_flax(self):
    shutil.rmtree(flags.FLAGS.test_tmpdir)
    batch_size = 8
    # Mujoco environment and associated demonstration dataset.
    environment = fakes.ContinuousEnvironment(
        episode_length=EPISODE_LENGTH,
        action_dim=CONTINUOUS_ACTION_DIM,
        observation_dim=CONTINUOUS_OBS_DIM,
        bounded=True)
    spec = specs.make_environment_spec(environment)

    networks = sac.make_networks(spec=spec)
    config = sac.SACConfig(batch_size=batch_size,
                           samples_per_insert_tolerance_rate=float('inf'),
                           min_replay_size=1)
    base_builder = sac.SACBuilder(config=config)
    direct_rl_batch_size = batch_size
    behavior_policy = sac.apply_policy_and_sample(networks)

    discriminator_module = DiscriminatorModule(spec, linen.Dense(1))

    def apply_fn(params: networks_lib.Params,
                 policy_params: networks_lib.Params, state: networks_lib.Params,
                 transitions: types.Transition, is_training: bool,
                 rng: networks_lib.PRNGKey) -> networks_lib.Logits:
      del policy_params
      variables = dict(params=params, **state)
      return discriminator_module.apply(
          variables,
          transitions.observation,
          transitions.action,
          transitions.next_observation,
          is_training=is_training,
          rng=rng,
          mutable=state.keys())

    def init_fn(rng):
      variables = discriminator_module.init(
          rng, dummy_obs, dummy_actions, dummy_obs, is_training=False, rng=rng)
      init_state, discriminator_params = variables.pop('params')
      return discriminator_params, init_state

    dummy_obs = utils.zeros_like(spec.observations)
    dummy_obs = utils.add_batch_dim(dummy_obs)
    dummy_actions = utils.zeros_like(spec.actions)
    dummy_actions = utils.add_batch_dim(dummy_actions)
    discriminator_network = networks_lib.FeedForwardNetwork(
        init=init_fn, apply=apply_fn)

    networks = ail.AILNetworks(discriminator_network, lambda x: x, networks)

    builder = ail.AILBuilder(
        base_builder,
        config=ail.AILConfig(
            is_sequence_based=False,
            share_iterator=True,
            direct_rl_batch_size=direct_rl_batch_size,
            discriminator_batch_size=2,
            policy_variable_name=None,
            min_replay_size=1),
        discriminator_loss=ail.losses.gail_loss(),
        make_demonstrations=fakes.transition_iterator(environment))

    counter = counting.Counter()
    # Construct the agent.
    agent = local_layout.LocalLayout(
        seed=0,
        environment_spec=spec,
        builder=builder,
        networks=networks,
        policy_network=behavior_policy,
        min_replay_size=1,
        batch_size=batch_size,
        counter=counter,
    )

    # Train the agent.
    train_loop = acme.EnvironmentLoop(environment, agent, counter=counter)
    train_loop.run(num_episodes=1)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Rescaling layers (e.g. to match action specs)."""

import dataclasses

from acme import specs
from jax import lax
import jax.numpy as jnp


@dataclasses.dataclass
class ClipToSpec:
  """Clips inputs to within a BoundedArraySpec."""
  spec: specs.BoundedArray

  def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
    return jnp.clip(inputs, self.spec.minimum, self.spec.maximum)


@dataclasses.dataclass
class RescaleToSpec:
  """Rescales inputs in [-1, 1] to match a BoundedArraySpec."""
  spec: specs.BoundedArray

  def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
    scale = self.spec.maximum - self.spec.minimum
    offset = self.spec.minimum
    inputs = 0.5 * (inputs + 1.0)  # [0, 1]
    output = inputs * scale + offset  # [minimum, maximum]
    return output


@dataclasses.dataclass
class TanhToSpec:
  """Squashes real-valued inputs to match a BoundedArraySpec."""
  spec: specs.BoundedArray

  def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
    scale = self.spec.maximum - self.spec.minimum
    offset = self.spec.minimum
    inputs = lax.tanh(inputs)  # [-1, 1]
    inputs = 0.5 * (inputs + 1.0)  # [0, 1]
    output = inputs * scale + offset  # [minimum, maximum]
    return output



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper that implements action repeats."""

from acme import types
from acme.wrappers import base
import dm_env


class ActionRepeatWrapper(base.EnvironmentWrapper):
  """Action repeat wrapper."""

  def __init__(self, environment: dm_env.Environment, num_repeats: int = 1):
    super().__init__(environment)
    self._num_repeats = num_repeats

  def step(self, action: types.NestedArray) -> dm_env.TimeStep:
    # Initialize accumulated reward and discount.
    reward = 0.
    discount = 1.

    # Step the environment by repeating action.
    for _ in range(self._num_repeats):
      timestep = self._environment.step(action)

      # Accumulate reward and discount.
      reward += timestep.reward * discount
      discount *= timestep.discount

      # Don't go over episode boundaries.
      if timestep.last():
        break

    # Replace the final timestep's reward and discount.
    return timestep._replace(reward=reward, discount=discount)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Loss function for IMPALA (Espeholt et al., 2018) [1].

[1] https://arxiv.org/abs/1802.01561
"""

from typing import Callable

from acme.agents.jax.impala import types
from acme.jax import utils
import haiku as hk
import jax.numpy as jnp
import numpy as np
import reverb
import rlax
import tree


def impala_loss(
    unroll_fn: types.PolicyValueFn,
    *,
    discount: float,
    max_abs_reward: float = np.inf,
    baseline_cost: float = 1.,
    entropy_cost: float = 0.,
) -> Callable[[hk.Params, reverb.ReplaySample], jnp.DeviceArray]:
  """Builds the standard entropy-regularised IMPALA loss function.

  Args:
    unroll_fn: A `hk.Transformed` object containing a callable which maps
      (params, observations_sequence, initial_state) -> ((logits, value), state)
    discount: The standard geometric discount rate to apply.
    max_abs_reward: Optional symmetric reward clipping to apply.
    baseline_cost: Weighting of the critic loss relative to the policy loss.
    entropy_cost: Weighting of the entropy regulariser relative to policy loss.

  Returns:
    A loss function with signature (params, data) -> loss_scalar.
  """

  def loss_fn(params: hk.Params,
              sample: reverb.ReplaySample) -> jnp.DeviceArray:
    """Batched, entropy-regularised actor-critic loss with V-trace."""

    # Extract the data.
    data = sample.data
    observations, actions, rewards, discounts, extra = (data.observation,
                                                        data.action,
                                                        data.reward,
                                                        data.discount,
                                                        data.extras)
    initial_state = tree.map_structure(lambda s: s[0], extra['core_state'])
    behaviour_logits = extra['logits']

    # Apply reward clipping.
    rewards = jnp.clip(rewards, -max_abs_reward, max_abs_reward)

    # Unroll current policy over observations.
    (logits, values), _ = unroll_fn(params, observations, initial_state)

    # Compute importance sampling weights: current policy / behavior policy.
    rhos = rlax.categorical_importance_sampling_ratios(logits[:-1],
                                                       behaviour_logits[:-1],
                                                       actions[:-1])

    # Critic loss.
    vtrace_returns = rlax.vtrace_td_error_and_advantage(
        v_tm1=values[:-1],
        v_t=values[1:],
        r_t=rewards[:-1],
        discount_t=discounts[:-1] * discount,
        rho_tm1=rhos)
    critic_loss = jnp.square(vtrace_returns.errors)

    # Policy gradient loss.
    policy_gradient_loss = rlax.policy_gradient_loss(
        logits_t=logits[:-1],
        a_t=actions[:-1],
        adv_t=vtrace_returns.pg_advantage,
        w_t=jnp.ones_like(rewards[:-1]))

    # Entropy regulariser.
    entropy_loss = rlax.entropy_loss(logits[:-1], jnp.ones_like(rewards[:-1]))

    # Combine weighted sum of actor & critic losses, averaged over the sequence.
    mean_loss = jnp.mean(policy_gradient_loss + baseline_cost * critic_loss +
                         entropy_cost * entropy_loss)  # []

    return mean_loss

  return utils.mapreduce(loss_fn, in_axes=(None, 0))



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adders for Reverb replay buffers."""

# pylint: disable=unused-import

from acme.adders.reverb.base import DEFAULT_PRIORITY_TABLE
from acme.adders.reverb.base import PriorityFn
from acme.adders.reverb.base import PriorityFnInput
from acme.adders.reverb.base import ReverbAdder
from acme.adders.reverb.base import Step

from acme.adders.reverb.episode import EpisodeAdder
from acme.adders.reverb.sequence import EndBehavior
from acme.adders.reverb.sequence import SequenceAdder
from acme.adders.reverb.transition import NStepTransitionAdder



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines distributed and local AIL agents, using JAX."""

import functools
from typing import Any, Callable, Iterator, Optional

from acme import specs
from acme import types
from acme.agents.jax import builders
from acme.agents.jax.ail import builder
from acme.agents.jax.ail import config as ail_config
from acme.agents.jax.ail import losses
from acme.agents.jax.ail import networks as ail_networks
from acme.jax import utils
from acme.jax.layouts import distributed_layout
from acme.jax.layouts import local_layout
from acme.utils import counting
from acme.utils import loggers
import dm_env


NetworkFactory = Callable[[specs.EnvironmentSpec], ail_networks.AILNetworks]


class DistributedAIL(distributed_layout.DistributedLayout):
  """Distributed program definition for AIL."""

  def __init__(self,
               environment_factory: Callable[[bool], dm_env.Environment],
               rl_agent: builders.GenericActorLearnerBuilder,
               config: ail_config.AILConfig,
               network_factory: NetworkFactory,
               seed: int,
               batch_size: int,
               make_demonstrations: Callable[[int], Iterator[types.Transition]],
               policy_network: Any,
               evaluator_policy_network: Any,
               num_actors: int,
               max_number_of_steps: Optional[int] = None,
               log_to_bigtable: bool = False,
               log_every: float = 10.0,
               prefetch_size: int = 4,
               discriminator_loss: Optional[losses.Loss] = None):
    assert discriminator_loss is not None
    logger_fn = functools.partial(
        loggers.make_default_logger,
        'learner',
        log_to_bigtable,
        time_delta=log_every,
        asynchronous=True,
        serialize_fn=utils.fetch_devicearray,
        steps_key='learner_steps')
    ail_builder = builder.AILBuilder(
        rl_agent=rl_agent,
        config=config,
        discriminator_loss=discriminator_loss,
        make_demonstrations=make_demonstrations,
        logger_fn=logger_fn)
    super().__init__(
        seed=seed,
        environment_factory=lambda: environment_factory(False),
        network_factory=network_factory,
        builder=ail_builder,
        policy_network=policy_network,
        evaluator_factories=[
            distributed_layout.default_evaluator(
                environment_factory=lambda: environment_factory(True),
                network_factory=network_factory,
                builder=ail_builder,
                policy_factory=evaluator_policy_network,
                log_to_bigtable=log_to_bigtable)
        ],
        num_actors=num_actors,
        max_number_of_steps=max_number_of_steps,
        prefetch_size=prefetch_size,
        log_to_bigtable=log_to_bigtable,
        actor_logger_fn=distributed_layout.get_default_logger_fn(
            log_to_bigtable, log_every),
    )


class AIL(local_layout.LocalLayout):
  """Local agent for AIL."""

  def __init__(self,
               spec: specs.EnvironmentSpec,
               rl_agent: builders.GenericActorLearnerBuilder,
               network: ail_networks.AILNetworks,
               config: ail_config.AILConfig,
               seed: int,
               batch_size: int,
               make_demonstrations: Callable[[int], Iterator[types.Transition]],
               policy_network: Any,
               samples_per_insert: float = 256,
               discriminator_loss: Optional[losses.Loss] = None,
               counter: Optional[counting.Counter] = None):
    self.builder = builder.AILBuilder(
        rl_agent=rl_agent,
        config=config,
        discriminator_loss=discriminator_loss,
        make_demonstrations=make_demonstrations)
    super().__init__(
        seed=seed,
        environment_spec=spec,
        builder=self.builder,
        networks=network,
        policy_network=policy_network,
        batch_size=batch_size,
        samples_per_insert=samples_per_insert,
        min_replay_size=config.min_replay_size,
        num_sgd_steps_per_step=config.num_sgd_steps_per_step,
        counter=counter,
    )



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AIL logits to AIL reward."""
from typing import Optional

from acme.agents.jax.ail import networks as ail_networks
from acme.jax import networks as networks_lib
import jax
import jax.numpy as jnp


def fairl_reward(
    max_reward_magnitude: Optional[float] = None
) -> ail_networks.ImitationRewardFn:
  """The FAIRL reward function (https://arxiv.org/pdf/1911.02256.pdf).

  Args:
    max_reward_magnitude: Clipping value for the reward.

  Returns:
    The function from logit to imitation reward.
  """

  def imitation_reward(logits: networks_lib.Logits) -> float:
    rewards = jnp.exp(jnp.clip(logits, a_max=20.)) * -logits
    if max_reward_magnitude is not None:
      # pylint: disable=invalid-unary-operand-type
      rewards = jnp.clip(
          rewards, a_min=-max_reward_magnitude, a_max=max_reward_magnitude)
    return rewards

  return imitation_reward


def gail_reward(
    reward_balance: float = .5,
    max_reward_magnitude: Optional[float] = None
) -> ail_networks.ImitationRewardFn:
  """GAIL reward function (https://arxiv.org/pdf/1606.03476.pdf).

  Args:
    reward_balance: 1 means log(D) reward, 0 means -log(1-D) and other values
      mean an average of the two.
    max_reward_magnitude: Clipping value for the reward.

  Returns:
    The function from logit to imitation reward.
  """

  def imitation_reward(logits: networks_lib.Logits) -> float:
    # Quick Maths:
    # logits = ln(D) - ln(1-D)
    # -softplus(-logits) = ln(D)
    # softplus(logits) = -ln(1-D)
    rewards = (
        reward_balance * -jax.nn.softplus(-logits) +
        (1 - reward_balance) * jax.nn.softplus(logits))
    if max_reward_magnitude is not None:
      # pylint: disable=invalid-unary-operand-type
      rewards = jnp.clip(
          rewards, a_min=-max_reward_magnitude, a_max=max_reward_magnitude)
    return rewards

  return imitation_reward



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SAC Builder."""
from typing import Callable, Iterator, List, Optional

import acme
from acme import adders
from acme import core
from acme import specs
from acme.adders import reverb as adders_reverb
from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax import actors
from acme.agents.jax import builders
from acme.agents.jax.sac import config as sac_config
from acme.agents.jax.sac import learning
from acme.agents.jax.sac import networks as sac_networks
from acme.datasets import reverb as datasets
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import optax
import reverb
from reverb import rate_limiters


class SACBuilder(builders.ActorLearnerBuilder):
  """SAC Builder."""

  def __init__(
      self,
      config: sac_config.SACConfig,
      logger_fn: Callable[[], loggers.Logger] = lambda: None,
  ):
    """Creates a SAC learner, a behavior policy and an eval actor.

    Args:
      config: a config with SAC hps
      logger_fn: a logger factory for the learner
    """
    self._config = config
    self._logger_fn = logger_fn

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: sac_networks.SACNetworks,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:
    # Create optimizers
    policy_optimizer = optax.adam(learning_rate=self._config.learning_rate)
    q_optimizer = optax.adam(learning_rate=self._config.learning_rate)

    return learning.SACLearner(
        networks=networks,
        tau=self._config.tau,
        discount=self._config.discount,
        entropy_coefficient=self._config.entropy_coefficient,
        target_entropy=self._config.target_entropy,
        rng=random_key,
        reward_scale=self._config.reward_scale,
        num_sgd_steps_per_step=self._config.num_sgd_steps_per_step,
        policy_optimizer=policy_optimizer,
        q_optimizer=q_optimizer,
        iterator=dataset,
        logger=self._logger_fn(),
        counter=counter)

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None) -> acme.Actor:
    assert variable_source is not None
    actor_core = actor_core_lib.batched_feed_forward_to_actor_core(
        policy_network)
    variable_client = variable_utils.VariableClient(variable_source, 'policy',
                                                    device='cpu')
    return actors.GenericActor(
        actor_core, random_key, variable_client, adder, backend='cpu')

  def make_replay_tables(
      self,
      environment_spec: specs.EnvironmentSpec,
  ) -> List[reverb.Table]:
    """Create tables to insert data into."""
    samples_per_insert_tolerance = (
        self._config.samples_per_insert_tolerance_rate *
        self._config.samples_per_insert)
    error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
    limiter = rate_limiters.SampleToInsertRatio(
        min_size_to_sample=self._config.min_replay_size,
        samples_per_insert=self._config.samples_per_insert,
        error_buffer=error_buffer)
    return [
        reverb.Table(
            name=self._config.replay_table_name,
            sampler=reverb.selectors.Uniform(),
            remover=reverb.selectors.Fifo(),
            max_size=self._config.max_replay_size,
            rate_limiter=limiter,
            signature=adders_reverb.NStepTransitionAdder.signature(
                environment_spec))
    ]

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    """Create a dataset iterator to use for learning/updating the agent."""
    dataset = datasets.make_reverb_dataset(
        table=self._config.replay_table_name,
        server_address=replay_client.server_address,
        batch_size=(
            self._config.batch_size * self._config.num_sgd_steps_per_step),
        prefetch_size=self._config.prefetch_size)
    return dataset.as_numpy_iterator()

  def make_adder(self,
                 replay_client: reverb.Client) -> Optional[adders.Adder]:
    """Create an adder which records data generated by the actor/environment."""
    return adders_reverb.NStepTransitionAdder(
        priority_fns={self._config.replay_table_name: None},
        client=replay_client,
        n_step=self._config.n_step,
        discount=self._config.discount)
